# Copyright 2023 Cisco Systems, Inc. and its affiliates

from pathlib import Path
from typing import ClassVar

from pydantic import ConfigDict

from catalystwan.api.templates.feature_template import FeatureTemplate


class CliTemplateModel(FeatureTemplate):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    config: str

    payload_path: ClassVar[Path] = Path(__file__).parent / "DEPRECATED"
    type: ClassVar[str] = "cli-template"



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from datetime import datetime
from typing import Dict, List, Optional, Union

from pydantic import BaseModel, ConfigDict, Field, IPvAnyAddress


class MTEdge(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    uuid: str
    configured_hostname: Optional[str] = Field(
        serialization_alias="configuredHostname", validation_alias="configuredHostname"
    )
    configured_system_ip: Optional[IPvAnyAddress] = Field(
        serialization_alias="configuredSystemIP", validation_alias="configuredSystemIP"
    )
    management_system_ip: Optional[IPvAnyAddress] = Field(
        serialization_alias="managementSystemIP", validation_alias="managementSystemIP"
    )
    device_model: Optional[str] = Field(serialization_alias="deviceModel", validation_alias="deviceModel")
    device_type: Optional[str] = Field(serialization_alias="deviceType", validation_alias="deviceType")


class TenantVPNMap(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    tenant_vpn: int = Field(serialization_alias="tenantVPN", validation_alias="tenantVPN")
    device_vpn: int = Field(serialization_alias="deviceVPN", validation_alias="deviceVPN")


class Tenant(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    name: str
    desc: str
    org_name: str = Field(serialization_alias="orgName", validation_alias="orgName")
    subdomain: str = Field(serialization_alias="subDomain", validation_alias="subDomain")
    flake_id: Optional[int] = Field(default=None, serialization_alias="flakeId", validation_alias="flakeId")
    vbond_address: Optional[str] = Field(
        default=None, serialization_alias="vBondAddress", validation_alias="vBondAddress"
    )
    edge_connector_system_ip: Optional[str] = Field(
        default=None, serialization_alias="edgeConnectorSystemIp", validation_alias="edgeConnectorSystemIp"
    )
    edge_connector_enable: Optional[bool] = Field(
        default=None, serialization_alias="edgeConnectorEnable", validation_alias="edgeConnectorEnable"
    )
    vsmarts: Optional[List[str]] = Field(default=None, serialization_alias="vSmarts", validation_alias="vSmarts")
    wan_edge_forecast: Optional[int] = Field(
        default=None, serialization_alias="wanEdgeForecast", validation_alias="wanEdgeForecast"
    )
    saml_sp_info: Optional[str] = Field(default=None, serialization_alias="samlSpInfo", validation_alias="samlSpInfo")
    idp_map: Union[Dict, str, None] = Field(default=None, serialization_alias="idpMap", validation_alias="idpMap")
    config_db_cluster_service_name: Optional[str] = Field(
        default=None, serialization_alias="configDBClusterServiceName", validation_alias="configDBClusterServiceName"
    )
    old_idp_map: Union[Dict, str, None] = Field(
        default=None, serialization_alias="oldIdpMap", validation_alias="oldIdpMap"
    )
    created_at: Optional[datetime] = Field(default=None, serialization_alias="createdAt", validation_alias="createdAt")
    rid: Optional[int] = Field(default=None, serialization_alias="@rid", validation_alias="@rid")
    edge_connector_tunnel_interface_name: Optional[str] = Field(
        default=None,
        serialization_alias="edgeConnectorTunnelInterfaceName",
        validation_alias="edgeConnectorTunnelInterfaceName",
    )
    tenant_id: Optional[str] = Field(default=None, serialization_alias="tenantId", validation_alias="tenantId")
    sp_metadata: Optional[str] = Field(default=None, serialization_alias="spMetadata", validation_alias="spMetadata")
    state: Optional[str] = None
    wan_edge_present: Optional[int] = Field(
        default=None, serialization_alias="wanEdgePresent", validation_alias="wanEdgePresent"
    )
    mt_edge: Optional[List[MTEdge]] = Field(default=None, serialization_alias="mtEdge", validation_alias="mtEdge")
    mt_edge_count: Optional[int] = Field(
        default=None, serialization_alias="mtEdgeCount", validation_alias="mtEdgeCount"
    )
    tenant_vpn_map: Optional[List[TenantVPNMap]] = Field(
        default=None, serialization_alias="tenantVPNmap", validation_alias="tenantVPNmap"
    )
    tenant_provider_vpn_count: Optional[int] = Field(
        default=None, serialization_alias="tenantProviderVPNCount", validation_alias="tenantProviderVPNCount"
    )


class TenantExport(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    name: str
    desc: str
    org_name: str = Field(serialization_alias="orgName", validation_alias="orgName")
    subdomain: str = Field(serialization_alias="subDomain", validation_alias="subDomain")
    wan_edge_forecast: Optional[int] = Field(
        default=None, serialization_alias="wanEdgeForecast", validation_alias="wanEdgeForecast"
    )
    is_destination_overlay_mt: Optional[bool] = Field(
        default=None,
        serialization_alias="isDestinationOverlayMT",
        validation_alias="isDestinationOverlayMT",
        description="required starting from 20.13",
    )
    migration_key: Optional[str] = Field(
        default=None,
        serialization_alias="migrationKey",
        validation_alias="migrationKey",
        pattern=r"^[a-zA-Z0-9]{8,32}$",
        description="required starting from 20.13",
    )



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from typing import List

from pydantic import AliasPath, BaseModel, ConfigDict, Field, field_validator

from catalystwan.api.configuration_groups.parcel import Global, _ParcelBase, as_global
from catalystwan.models.common import TLOCColor


class AppProbeMapItem(BaseModel):
    color: Global[TLOCColor]
    dscp: Global[int]

    @field_validator("dscp")
    @classmethod
    def check_rate(cls, dscp: Global):
        assert 0 <= int(dscp.value) <= 63
        return dscp


class AppProbeEntry(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    map: List[AppProbeMapItem] = Field(default=[])
    forwarding_class_name: Global[str] = Field(
        serialization_alias="forwardingClass", validation_alias="forwardingClass"
    )


class AppProbeParcel(_ParcelBase):
    entries: List[AppProbeEntry] = Field(default=[], validation_alias=AliasPath("data", "entries"))

    def add_fowarding_class(self, forwarding_class_name: str):
        self.entries.append(
            AppProbeEntry(
                forwarding_class_name=as_global(forwarding_class_name),
            )
        )

    def add_map(self, color: TLOCColor, dscp: int):
        entry = self.entries[0]
        entry.map.append(AppProbeMapItem(color=as_global(color, TLOCColor), dscp=as_global(dscp)))



# Copyright 2023 Cisco Systems, Inc. and its affiliates

# mypy: disable-error-code="empty-body"
from enum import Enum
from typing import List

from pydantic import BaseModel, Field, RootModel

from catalystwan.endpoints import APIEndpoints, get, put
from catalystwan.typed_list import DataSequence


class StatusEnum(str, Enum):
    enable = "enable"
    disable = "disable"
    bypass = "bypass"
    custom = "custom"


index_name_field = Field(..., alias="indexName", pattern="^[a-z]+$")


class Status(BaseModel):
    index_name: str = index_name_field
    status: StatusEnum
    display_name: str = Field(alias="displayName")


class UpdateStatus(BaseModel):
    index_name: str = index_name_field
    status: StatusEnum


class EnabledIndex(BaseModel):
    index_name: str = index_name_field


SingleList = RootModel[List[str]]  # trick to make endpoint return list of strings


class DisabledDeviceList(SingleList):
    pass


class DisabledDeviceListResponse(SingleList):
    pass


class UpdateIndexResponse(BaseModel):
    response: bool


class EnabledIndexDeviceListResponse(SingleList):
    pass


class MonitoringStatus(APIEndpoints):
    @get("/statistics/settings/disable/devicelist/{indexName}")
    def get_disabled_device_list(self, indexName: str) -> DisabledDeviceListResponse:
        ...

    @put("/statistics/settings/disable/devicelist/{indexName}", "response")
    def update_statistics_device_list(self, indexName: str, payload: DisabledDeviceList) -> UpdateIndexResponse:
        ...

    @get("/statistics/settings/status")
    def get_statistics_settings(self) -> DataSequence[Status]:
        ...

    @put("/statistics/settings/status")
    def update_statistics_settings(self, payload: List[UpdateStatus]) -> DataSequence[Status]:
        ...

    @get("/statistics/settings/status/device")
    def get_enabled_index_for_device(self, params: dict) -> EnabledIndexDeviceListResponse:
        ...



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from __future__ import annotations

import logging
from datetime import datetime
from pathlib import Path
from typing import Dict, List

from packaging.version import Version  # type: ignore

from catalystwan.api.tenant_migration_api import TenantMigrationAPI
from catalystwan.endpoints.troubleshooting_tools.device_connectivity import NPingRequest
from catalystwan.exceptions import TenantMigrationPreconditionsError
from catalystwan.models.tenant import TenantExport
from catalystwan.session import ManagerSession, create_manager_session
from catalystwan.utils.personality import Personality
from catalystwan.utils.session_type import SessionType

logger = logging.getLogger(__name__)


def raise_or_log_precondition_check(msg: str, raises: bool) -> None:
    if raises:
        raise TenantMigrationPreconditionsError(msg)
    logger.warning(msg)


def check_control_connectivity_from_edge_devices(session: ManagerSession, host: str, attempts: int = 2) -> bool:
    """Checks that all edge devices can reach specified host using ping on transport VPN 0 which carries control traffic

    Args:
        session (ManagerSession): Session logged as device owner user
        host (str): IP address or domain name to ping to
        attempts (int, optional): Number of attempts. Defaults to 2.

    Returns:
        bool: True only if all devices can reach specified host
    """
    conn_status_by_device_id: Dict[str, bool] = {}
    edge_devices = session.api.devices.get_reachable_devices(personality=Personality.EDGE)
    if not edge_devices:
        logger.warning("No edge devices found to perform connectivity check!")
        return False
    for edge in edge_devices:
        logger.info(f"Checking {edge.id} can reach {host}...")
        conn_status_by_device_id[edge.id] = False
        for _ in range(attempts):
            ping_request = NPingRequest(host=host, vpn="0")
            ping_result = session.endpoints.troubleshooting_tools.device_connectivity.nping_device(
                edge.id, ping_request
            )
            if ping_result.loss_percentage < 50.0:
                conn_status_by_device_id[edge.id] = True
                break
    logger.info(f"Connectivity status to host: {host} by device:\n{conn_status_by_device_id}")
    return all(conn_status_by_device_id.values())


def migration_preconditions_check(
    origin_session: ManagerSession,
    target_session: ManagerSession,
    tenant: TenantExport,
    validator: str,
    raises: bool,
) -> bool:
    """Perform precondition checks prior tenant migration

    Args:
        origin_session (ManagerSession): session to migration origin
        target_session (ManagerSession): session to migration target
        tenant (MigrationTenant): Tenant object containig required fields: desc, name, subdomain, org_name
        validator (str): Target Validator (VBOND) IP address or domain name
        raises (bool): When true precondition check will raise, when false only warning will be logged

    Returns:
        bool: true only when all preconditions pass
    """
    problems: List[str] = []
    # Check if both platforms uses same build version
    logger.info("Checking if both platform versions match...")
    if target_session._platform_version != origin_session._platform_version:
        problems.append(
            f"Migration source and target expect to have same platform version but found "
            f"origin: {origin_session._platform_version} "
            f"target: {target_session._platform_version}",
        )
    # Export Params check
    logger.info("Performing export parameters checks...")
    if origin_session.api_version >= Version("20.13"):
        if tenant.is_destination_overlay_mt is None or tenant.migration_key is None:
            problems.append("'isDestinationOverlayMT' and 'migrationKey' must be provided using >= 20.13")
    # Target checks
    logger.info("Performing target checks...")
    target_org = target_session.endpoints.configuration_settings.get_organizations().first().org
    # Checks for MT target
    if target_session.session_type == SessionType.PROVIDER:
        if origin_session.api_version >= Version("20.13"):
            if not tenant.is_destination_overlay_mt:
                problems.append("Migrating to MT using >= 20.13 but 'isDestinationOverlayMT' is not set")
        if origin_session.session_type != SessionType.SINGLE_TENANT:
            problems.append(
                "Migration to MT (using provider) is expected to be initiated from ST (using single tenant)"
            )
        if not tenant.org_name.startswith(target_org):
            problems.append(f"Provided '{tenant.org_name}' but target organization is '{target_org}'")

    # Checks for ST target
    elif target_session.session_type == SessionType.SINGLE_TENANT:
        if target_session.api_version < Version("20.13"):
            problems.append("Migration to ST not supported prior 20.13")
        if tenant.is_destination_overlay_mt is True:
            problems.append("Migrating to ST but 'isDestinationOverlayMT' is set to True")
        if origin_session.session_type != SessionType.PROVIDER:
            problems.append(
                "Migration to ST (using single tenant) is expected to be initiated from MT (using provider)"
            )
        if tenant.org_name != target_org:
            problems.append(f"Provided '{tenant.org_name}' but target organization is '{target_org}'")
    else:
        problems.append(
            f"Migration target is expected to be executed as single tenant or provider "
            f"but found: {target_session.session_type}"
        )

    # Check if migrated devices can access target VBOND/Validator
    logger.info("Checking if migrated devices can reach target validator...")
    conn_check = False
    if origin_session.session_type == SessionType.PROVIDER:
        with create_manager_session(
            url=origin_session.url,
            username=origin_session.username,
            password=origin_session.password,
            port=origin_session.port,
            subdomain=tenant.subdomain,
            logger=origin_session.logger,
        ) as provider_as_tenant_session:
            conn_check = check_control_connectivity_from_edge_devices(provider_as_tenant_session, validator)
    else:
        conn_check = check_control_connectivity_from_edge_devices(origin_session, validator)
    if not conn_check:
        problems.append(f"Migrated devices cannot reach validator: {validator}")

    # Generate error message
    if problems:
        problem_lines = "\n".join(problems)
        message = f"Found {len(problems)} problems in precondition check for migration:\n{problem_lines}"
        raise_or_log_precondition_check(message, raises)
        return False
    logger.info("Preconditions checks for tenant migration succeeded!")
    return True


def migration_workflow(
    origin_session: ManagerSession,
    target_session: ManagerSession,
    workdir: Path,
    tenant: TenantExport,
    validator: str,
    raises: bool = True,
):
    """Performs migration from origin sdwan instance to target sdwan instance based on:
    https://www.cisco.com/c/en/us/td/docs/routers/sdwan/configuration/system-interface/vedge-20-x/systems-interfaces-book/sdwan-multitenancy.html#concept_sjj_jmm_z4b
    0. Perform pre-condition checks
    1. Export the deployment and configuration data from origin Cisco vManage instance controlling the overlay.
    2. Check the status of the data export task in Cisco vManage. Download the data when the task succeeds.
    3. On a target Cisco vManage instance, import the data exported from the origin overlay.
    4. Collect migration token using the token URL obtained in response to the API call in Step 3.
    5. On origin Cisco vManage instance, initiate the migration of the overlay to target deployment.

    Args:
        origin_session (ManagerSession): session to migration origin
        target_session (ManagerSession): session to migration target
        workdir (Path): directory to store migration artifacts (token and export file)
        tenant (MigrationTenant): Tenant object containig required fields: desc, name, subdomain, org_name
        validator (str): Target Validator (VBOND) IP address or domain name
        raises (bool): When true precondition check will raise, when false only warning will be logged
    """
    workdir.mkdir(parents=True, exist_ok=True)
    logger.info("0/5 Performing pre-checks ...")
    migration_preconditions_check(origin_session, target_session, tenant, validator, raises)
    origin_api = TenantMigrationAPI(origin_session)
    target_api = TenantMigrationAPI(target_session)
    migration_timestamp = datetime.now().strftime("%Y%m%d%H%M")
    migration_file_prefix = f"{tenant.name}-{origin_api.session.server_name}-{migration_timestamp}"
    export_path = workdir / f"{migration_file_prefix}.tar.gz"
    token_path = workdir / f"{migration_file_prefix}.token"

    logger.info(f"1/5 Exporting {tenant.name} ...")
    export_task = origin_api.export_tenant(tenant=tenant)
    remote_filename = export_task.wait_for_file()

    logger.info(f"2/5 Downloading {remote_filename} to {export_path} ...")
    origin_api.download(export_path, remote_filename)

    logger.info(f"3/5 Importing {export_path} ...")
    import_task = target_api.import_tenant(export_path, tenant.migration_key)

    logger.info("4/5 Obtaining migration token ...")
    import_task.wait_for_completed()
    migration_id = import_task.import_info.migration_token_query_params.migration_id
    target_api.store_token(migration_id, token_path)

    logger.info(f"5/5 Initiating network migration: {migration_id}, using token file: {token_path} ...")
    migrate_task = origin_api.migrate_network(token_path)
    migrate_task.wait_for_completed()
    logger.info(f"5/5 {tenant.name} migration completed successfully!")



# Copyright 2023 Cisco Systems, Inc. and its affiliates

# mypy: disable-error-code="empty-body"
from datetime import datetime
from typing import Dict, List, Optional, Set

from pydantic import BaseModel, ConfigDict, Field

from catalystwan.endpoints import APIEndpoints, delete, get, post, put, versions
from catalystwan.typed_list import DataSequence


class User(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    username: str = Field(serialization_alias="userName", validation_alias="userName")
    password: Optional[str] = None
    group: List[str]
    locale: Optional[str] = None
    description: Optional[str] = None
    resource_group: Optional[str] = Field(
        default=None,
        serialization_alias="resGroupName",
        validation_alias="resGroupName",
        description="can be set only for >=20.5, <20.13 where default value is 'global'",
    )
    resource_domain: Optional[str] = Field(
        default=None,
        serialization_alias="resourceDomainName",
        validation_alias="resourceDomainName",
        description="can be set only for >=20.13 where default value is 'all'",
    )


class UserUpdateRequest(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    username: str = Field(serialization_alias="userName", validation_alias="userName")
    current_password: bool = Field(
        serialization_alias="currentPassword", validation_alias="currentPassword", default=False
    )
    show_password: bool = Field(serialization_alias="showPassword", validation_alias="showPassword", default=False)
    show_confirm_password: bool = Field(
        serialization_alias="showConfirmPassword", validation_alias="showConfirmPassword", default=False
    )
    current_user_password: Optional[str] = Field(
        default=None,
        serialization_alias="currentUserPassword",
        validation_alias="currentUserPassword",
    )
    password: Optional[str] = None
    group: Optional[List[str]] = None
    locale: Optional[str] = None
    description: Optional[str] = None
    resource_group: Optional[str] = Field(
        default=None,
        serialization_alias="resGroupName",
        validation_alias="resGroupName",
        description="can be set only for >=20.5, <20.13 where default value is 'global'",
    )
    resource_domain: Optional[str] = Field(
        default=None,
        serialization_alias="resourceDomainName",
        validation_alias="resourceDomainName",
        description="can be set only for >=20.13 where default value is 'all'",
    )


class UserRole(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    is_admin: bool = Field(serialization_alias="isAdmin", validation_alias="isAdmin")


class UserAuthType(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    user_auth_type: str = Field(serialization_alias="userAuthType", validation_alias="userAuthType")


class UserGroupTask(BaseModel):
    enabled: bool
    feature: str
    read: bool
    write: bool


class UserGroup(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    group_name: str = Field(serialization_alias="groupName", validation_alias="groupName")
    tasks: List[UserGroupTask]

    def __map_task_index_by_feature(self) -> Dict[str, int]:
        """Returns a mapping of internal task list index by feature name the task contains"""
        return {task.feature: index for index, task in enumerate(self.tasks)}

    def get_task(self, feature: str) -> Optional[UserGroupTask]:
        """Returns task object by feature name if found"""
        if index := self.__map_task_index_by_feature().get(feature, None) is not None:
            return self.tasks[index]
        return None

    def update_task(self, task: UserGroupTask):
        """Updates existing task entry or creates new entry if not exist"""
        if index := self.__map_task_index_by_feature().get(task.feature, None) is not None:
            self.tasks[index] = task
        else:
            self.tasks.append(task)

    def enable_read(self, features: Set[str]):
        """Enables read-only permissions for given feature set"""
        for feature in features:
            self.update_task(UserGroupTask(enabled=True, feature=feature, read=True, write=False))

    def enable_read_and_write(self, features: Set[str]):
        """Enables read-write permissions for given feature set"""
        for feature in features:
            self.update_task(UserGroupTask(enabled=True, feature=feature, read=True, write=True))

    def disable(self, features: Set[str]):
        """Disables access permissions for given feature set"""
        for feature in features:
            self.update_task(UserGroupTask(enabled=False, feature=feature, read=False, write=False))


class UserResetRequest(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    username: str = Field(serialization_alias="userName", validation_alias="userName")


class ActiveSession(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    uuid: str
    source_ip: Optional[str] = Field(default=None, serialization_alias="sourceIp", validation_alias="sourceIp")
    remote_host: Optional[str] = Field(default=None, serialization_alias="remoteHost", validation_alias="remoteHost")
    raw_username: Optional[str] = Field(default=None, serialization_alias="rawUserName", validation_alias="rawUserName")
    raw_id: Optional[str] = Field(default=None, serialization_alias="rawId", validation_alias="rawId")
    tenant_domain: Optional[str] = Field(
        default=None, serialization_alias="tenantDomain", validation_alias="tenantDomain"
    )
    user_group: Optional[str] = Field(
        default=None, serialization_alias="userGroup", validation_alias="userGroup"
    )  # workaround: should be List[str] but JSON array is quoted in response
    user_mode: Optional[str] = Field(default=None, serialization_alias="userMode", validation_alias="userMode")
    create_date_time: Optional[datetime] = Field(
        default=None, serialization_alias="createDateTime", validation_alias="createDateTime"
    )
    tenant_id: Optional[str] = Field(default=None, serialization_alias="tenantId", validation_alias="tenantId")
    last_accessed_time: Optional[datetime] = Field(
        default=None, serialization_alias="lastAccessedTime", validation_alias="lastAccessedTime"
    )


class SessionsDeleteRequest(BaseModel):
    data: List[ActiveSession]

    @classmethod
    def from_active_session_list(cls, sessions: List[ActiveSession]) -> "SessionsDeleteRequest":
        sessions_delete_request = SessionsDeleteRequest(data=[])
        for session in sessions:
            sessions_delete_request.data.append(
                ActiveSession(uuid=session.uuid, tenantId=session.tenant_id, rawId=session.raw_id)  # type: ignore
            )
        return sessions_delete_request


class InvalidateSessionMessage(BaseModel):
    message: Optional[str] = None


class ProfilePasswordUpdateRequest(BaseModel):
    oldpassword: str
    newpassword: str


class ResourceGroup(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    id: Optional[str] = None
    name: str
    desc: str
    site_ids: List[int] = Field(serialization_alias="siteIds", validation_alias="siteIds")
    device_ips: Optional[List[str]] = Field(default=None, serialization_alias="deviceIPs", validation_alias="deviceIPs")
    mgmt_sytem_ips_map: Optional[Dict[str, str]] = Field(
        default=None, serialization_alias="mgmtSytemIpsMap", validation_alias="mgmtSytemIpsMap"
    )
    uuid_sytem_ips_map: Optional[Dict[str, str]] = Field(
        default=None, serialization_alias="uuidSytemIpsMap", validation_alias="uuidSytemIpsMap"
    )


class ResourceGroupUpdateRequest(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    id: str
    name: str
    desc: str
    site_ids: List[int] = Field(serialization_alias="siteIds", validation_alias="siteIds")


class ResourceGroupSwitchRequest(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    resource_group_name: str = Field(serialization_alias="resourceGroupName", validation_alias="resourceGroupName")


class AdministrationUserAndGroup(APIEndpoints):
    def create_colo_group(self):
        # POST /admin/cologroup
        ...

    def create_group_grid_columns(self):
        # GET /admin/usergroup/definition
        ...

    @post("/admin/user")
    def create_user(self, payload: User) -> None:
        ...

    @post("/admin/usergroup")
    def create_user_group(self, payload: UserGroup) -> None:
        ...

    def create_vpn_group(self):
        # POST /admin/vpngroup
        ...

    def delete_colo_group(self):
        # DELETE /admin/cologroup/{id}
        ...

    @delete("/admin/user/{username}")
    def delete_user(self, username: str) -> None:
        ...

    @delete("/admin/usergroup/{group_name}")
    def delete_user_group(self, group_name: str) -> None:
        ...

    def delete_vpn_group(self):
        # DELETE /admin/vpngroup/{id}
        ...

    def edit_colo_group(self):
        # PUT /admin/cologroup/{id}
        ...

    def edit_vpn_group(self):
        # PUT /admin/vpngroup/{id}
        ...

    @get("/admin/user/userAuthType")
    def find_user_auth_type(self) -> UserAuthType:
        ...

    @get("/admin/usergroup", "data")
    def find_user_groups(self) -> DataSequence[UserGroup]:
        ...

    def find_user_groups_as_key_value(self):
        # GET /admin/usergroup/keyvalue
        ...

    @get("/admin/user/role")
    def find_user_role(self) -> UserRole:
        ...

    @get("/admin/user", "data")
    def find_users(self) -> DataSequence[User]:
        ...

    @get("/admin/user/activeSessions", "data")
    def get_active_sessions(self) -> DataSequence[ActiveSession]:
        ...

    def get_colo_groups(self):
        # GET /admin/cologroup
        ...

    def get_vpn_groups(self):
        # GET /admin/vpngroup
        ...

    @delete("/admin/user/removeSessions", "data")
    def remove_sessions(self, payload: SessionsDeleteRequest) -> InvalidateSessionMessage:
        ...

    @post("/admin/user/reset")
    def reset_user(self, payload: UserResetRequest) -> None:
        ...

    @versions(">20.4, <20.13")
    @get("/admin/resourcegroup")
    def find_resource_groups(self) -> DataSequence[ResourceGroup]:
        ...

    @versions(">20.4, <20.13")
    @post("/admin/resourcegroup/switch")
    def switch_resource_group(self, payload: ResourceGroupSwitchRequest) -> None:
        ...

    @versions(">20.4, <20.13")
    @put("/admin/resourcegroup/{group_id}")
    def update_resource_group(self, group_id: str, payload: ResourceGroupUpdateRequest) -> None:
        ...

    @versions(">20.4, <20.13")
    @delete("/admin/resourcegroup/{group_id}", json={})
    def delete_resource_group(self, group_id: str) -> None:
        ...

    @versions(">20.4, <20.13")
    @post("/admin/resourcegroup")
    def create_resource_group(self, payload: ResourceGroup) -> None:
        ...

    @versions(">20.4, <20.13")
    def resource_group_name(self):
        # GET /admin/user/resourceGroupName
        ...

    def update_admin_password(self):
        # POST /admin/user/admin/password
        ...

    @put("/admin/user/password/{username}")
    def update_password(self, username: str, payload: UserUpdateRequest) -> None:
        ...

    def update_profile_locale(self):
        # PUT /admin/user/profile/locale
        ...

    @put("/admin/user/profile/password")
    def update_profile_password(self, payload: ProfilePasswordUpdateRequest) -> None:
        ...

    @put("/admin/user/{username}")
    def update_user(self, username: str, payload: UserUpdateRequest) -> None:
        ...

    @put("/admin/usergroup/{group_name}")
    def update_user_group(self, group_name: str, payload: UserGroup) -> None:
        ...

    def validate_password(self):
        # POST /admin/user/password/validate
        ...



# Copyright 2024 Cisco Systems, Inc. and its affiliates

import re
from typing import Optional
from uuid import UUID

from pydantic import BaseModel, ConfigDict, Field, model_validator


class BoostrapConfigurationDetails(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    bootstrap_config: str = Field(serialization_alias="bootstrapConfig", validation_alias="bootstrapConfig")

    uuid: Optional[UUID] = Field(default=None)
    otp: Optional[str] = Field(default=None)
    vbond: Optional[str] = Field(default=None)
    org: Optional[str] = Field(default=None)

    @model_validator(mode="after")
    def try_parse(self):
        matched_groups = re.search(
            r"- uuid : (\S+).*?- otp : (\S+).*?- vbond : (\S+).*?- org : (\S+)", self.bootstrap_config, re.DOTALL
        )
        if matched_groups:
            self.uuid = matched_groups.group(1)
            self.otp = matched_groups.group(2)
            self.vbond = matched_groups.group(3)
            self.org = matched_groups.group(4)
        return self



# Copyright 2023 Cisco Systems, Inc. and its affiliates

# mypy: disable-error-code="empty-body"
from enum import Enum
from typing import List, Optional

from pydantic import BaseModel, ConfigDict, Field

from catalystwan.endpoints import APIEndpoints, post


class PingProbeType(str, Enum):
    ICMP = "icmp"
    TCP = "tcp"
    UDP = "udp"


class NPingRequest(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    host: str
    vpn: str
    probe_type: PingProbeType = Field(default=PingProbeType.ICMP, serialization_alias="probeType")
    count: Optional[str] = None
    dest_port: Optional[str] = Field(default=None, serialization_alias="destPort")
    df: Optional[str] = None
    interface_ip: Optional[str] = Field(default=None, serialization_alias="interfaceIP")
    mtu: Optional[str] = None
    rapid: Optional[str] = None
    size: Optional[str] = None
    source: Optional[str] = None
    source_port: Optional[str] = Field(default=None, serialization_alias="sourcePort")
    tos: Optional[str] = None
    ttl: Optional[str] = None


class NPingResult(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    raw_output: List[str] = Field(validation_alias="rawOutput")
    packets_transmitted: int = Field(validation_alias="packetsTransmitted")
    packets_received: int = Field(validation_alias="packetsReceived")
    loss_percentage: float = Field(validation_alias="lossPercentage")
    min_round_trip: float = Field(validation_alias="minRoundTrip")
    max_round_trip: float = Field(validation_alias="maxRoundTrip")
    avg_round_trip: float = Field(validation_alias="avgRoundTrip")


class TroubleshootingToolsDeviceConnectivity(APIEndpoints):
    def copy_admin_tech_on_device(self):
        # POST /device/tools/admintech/copy
        ...

    def create_admin_tech(self):
        # POST /device/tools/admintech
        ...

    def delete_admin_tech_file(self):
        # DELETE /device/tools/admintech/{requestID}
        ...

    def delete_admin_tech_on_device(self):
        # DELETE /device/tools/admintech/delete
        ...

    def download_admin_tech_file(self):
        # GET /device/tools/admintech/download/{filename}
        ...

    def factory_reset(self):
        # POST /device/tools/factoryreset
        ...

    def get_control_connections(self):
        # GET /troubleshooting/control/{uuid}
        ...

    def get_device_configuration(self):
        # GET /troubleshooting/devicebringup
        ...

    def get_in_progress_count(self):
        # GET /device/tools/admintechs/inprogress
        ...

    def list_admin_techs(self):
        # GET /device/tools/admintechs
        ...

    def list_admin_techs_on_device(self):
        # POST /device/tools/admintechlist
        ...

    @post("/device/tools/nping/{device_ip}")
    def nping_device(self, device_ip: str, payload: NPingRequest) -> NPingResult:
        ...

    def ping_device(self):
        # POST /device/tools/ping/{deviceIP}
        ...

    def process_interface_reset(self):
        # POST /device/tools/reset/interface/{deviceIP}
        ...

    def process_port_hop_color(self):
        # POST /device/tools/porthopcolor/{deviceIP}
        ...

    def process_reset_user(self):
        # POST /device/tools/resetuser/{deviceIP}
        ...

    def service_path(self):
        # POST /device/tools/servicepath/{deviceIP}
        ...

    def traceroute_device(self):
        # POST /device/tools/traceroute/{deviceIP}
        ...

    def tunnel_path(self):
        # POST /device/tools/tunnelpath/{deviceIP}
        ...

    def upload_admin_tech(self):
        # POST /device/tools/admintechs/upload
        ...



# Copyright 2024 Cisco Systems, Inc. and its affiliates

# type: ignore
import unittest
from unittest.mock import MagicMock, patch

from catalystwan.api.monitoring_status_api import MonitoringStatusAPI
from catalystwan.endpoints.monitoring.status import (
    DisabledDeviceListResponse,
    EnabledIndexDeviceListResponse,
    Status,
    StatusEnum,
    UpdateIndexResponse,
    UpdateStatus,
)
from catalystwan.typed_list import DataSequence


class TestMonitoringStatusAPI(unittest.TestCase):
    def test_get_statistic_settings(self):
        # Arrange
        self.api._endpoints.get_statistics_settings.return_value = self.returned_get_statistic_settings
        # Act
        result = self.api.get_statistic_settings()
        # Assert
        assert result == self.expected_get_statistic_settings

    def test_update_statistics_settings(self):
        # Arrange
        self.api._endpoints.update_statistics_settings.return_value = self.returned_update_statistics_settings
        # Act
        result = self.api.update_statistics_settings(self.update_statistics_settings_payload)
        # Assert
        assert result == self.expected_update_statistics_settings

    def test_get_disabled_devices_by_index(self):
        # Arrange
        self.api._endpoints.get_disabled_device_list.return_value = self.returned_list_disabled_devices
        # Act
        result = self.api.get_disabled_devices_by_index(self.indexName)
        # Assert
        assert result == self.expected_list_disabled_devices

    def test_update_disabled_devices_by_index(self):
        # Arrange
        self.api._endpoints.update_statistics_device_list.return_value = self.returned_update_list_disabled_devices
        # Act
        result = self.api.update_disabled_devices_by_index(self.indexName, self.update_disabled_devices_by_index_list)
        # Assert
        assert result == self.expected_update_list_disabled_devices

    def test_get_enabled_index_for_device(self):
        # Arrange
        self.api._endpoints.get_enabled_index_for_device.return_value = self.returned_list_enabled_indexes
        # Act
        result = self.api.get_enabled_index_for_device(self.device_id)
        # Assert
        assert result == self.expected_list_enabled_indexes

    @patch("catalystwan.session.ManagerSession")
    def setUp(self, mock_session) -> None:
        self.session = mock_session
        self.session.api_version = None
        self.session.session_type = None
        self.session.password = "P4s$w0rD"  # pragma: allowlist secret
        self.api = MonitoringStatusAPI(self.session)
        self.api._endpoints = MagicMock()
        self.device_id = "1.1.1.1"
        self.indexName = "aggregatedappsdpisummary"
        self.get_statistic_settings = [
            Status(indexName="interfacestatistics", status=StatusEnum.enable, displayName="Interface Statistics"),
            Status(indexName="eioltestatistics", status=StatusEnum.enable, displayName="EIO LTE Statistics"),
            Status(indexName="sulstatistics", status=StatusEnum.enable, displayName="SUL Statistics"),
        ]
        self.returned_get_statistic_settings = DataSequence(Status, self.get_statistic_settings)
        self.expected_get_statistic_settings = DataSequence(Status, self.get_statistic_settings)
        self.update_statistics_settings_payload = [
            UpdateStatus(indexName="interfacestatistics", status=StatusEnum.disable),
            UpdateStatus(indexName="eioltestatistics", status=StatusEnum.disable),
            UpdateStatus(indexName="sulstatistics", status=StatusEnum.disable),
        ]
        self.update_statistics_settings = [
            Status(indexName="interfacestatistics", status=StatusEnum.disable, displayName="Interface Statistics"),
            Status(indexName="eioltestatistics", status=StatusEnum.disable, displayName="EIO LTE Statistics"),
            Status(indexName="sulstatistics", status=StatusEnum.disable, displayName="SUL Statistics"),
        ]
        self.returned_update_statistics_settings = DataSequence(Status, self.update_statistics_settings)
        self.expected_update_statistics_settings = DataSequence(Status, self.update_statistics_settings)
        self.update_disabled_devices_by_index_list = ["127.0.0.1", "10.0.1.0"]
        self.expected_update_list_disabled_devices = True
        self.returned_update_list_disabled_devices = UpdateIndexResponse(response=True)
        self.returned_list_disabled_devices = DisabledDeviceListResponse(["1.1.1.1", "2.2.2.2", "3.3.3.3"])
        self.expected_list_disabled_devices = ["1.1.1.1", "2.2.2.2", "3.3.3.3"]
        self.expected_list_enabled_indexes = [
            "interfacestatistics",
            "eioltestatistics",
            "sulstatistics",
            "qosstatistics",
            "devicesystemstatusstatistics",
            "bridgemacstatistics",
            "dpistatistics",
            "approutestatsstatistics",
            "wlanclientinfostatistics",
            "flowlogstatistics",
            "urlf",
            "fwall",
            "vnfstatistics",
            "umbrella",
            "apphostingstatistics",
            "utddaqioxstatistics",
            "ipsalert",
            "aggregatedappsdpistatistics",
            "bridgeinterfacestatistics",
            "artstatistics",
            "trackerstatistics",
            "cloudxstatistics",
        ]
        self.returned_list_enabled_indexes = EnabledIndexDeviceListResponse(self.expected_list_enabled_indexes)



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from typing import List, Mapping, Union

from pydantic import Field
from typing_extensions import Annotated

from .policy.app_probe import AppProbeMapItem, AppProbeParcel
from .policy.application_list import ApplicationFamilyListEntry, ApplicationListEntry, ApplicationListParcel
from .policy.color_list import ColorEntry, ColorParcel
from .policy.data_prefix import DataPrefixEntry, DataPrefixParcel
from .policy.expanded_community_list import ExpandedCommunityParcel
from .policy.fowarding_class import FowardingClassParcel, FowardingClassQueueEntry
from .policy.ipv6_data_prefix import IPv6DataPrefixEntry, IPv6DataPrefixParcel
from .policy.ipv6_prefix_list import IPv6PrefixListEntry, IPv6PrefixListParcel
from .policy.policier import PolicierEntry, PolicierParcel
from .policy.prefered_group_color import Preference, PreferredColorGroupEntry, PreferredColorGroupParcel
from .policy.prefix_list import PrefixListEntry, PrefixListParcel
from .policy.sla_class import FallbackBestTunnel, SLAAppProbeClass, SLAClassCriteria, SLAClassListEntry, SLAClassParcel
from .policy.standard_community import StandardCommunityEntry, StandardCommunityParcel
from .policy.tloc_list import TlocEntry, TlocParcel
from .security.application_list import (
    SecurityApplicationFamilyListEntry,
    SecurityApplicationListEntry,
    SecurityApplicationListParcel,
)
from .security.data_prefix import SecurityDataPrefixEntry, SecurityDataPrefixParcel
from .security.fqdn import FQDNDomainParcel, FQDNListEntry
from .security.geolocation_list import GeoLocationListEntry, GeoLocationListParcel
from .security.ips_signature import IPSSignatureListEntry, IPSSignatureParcel
from .security.local_domain import LocalDomainListEntry, LocalDomainParcel
from .security.protocol_list import ProtocolListEntry, ProtocolListParcel
from .security.security_port import SecurityPortListEntry, SecurityPortParcel
from .security.url import BaseURLListEntry, URLAllowParcel, URLBlockParcel
from .security.zone import SecurityZoneListEntry, SecurityZoneListParcel

AnyPolicyObjectParcel = Annotated[
    Union[
        AppProbeParcel,
        ApplicationListParcel,
        ColorParcel,
        DataPrefixParcel,
        ExpandedCommunityParcel,
        FowardingClassParcel,
        IPv6DataPrefixParcel,
        IPv6PrefixListParcel,
        PrefixListParcel,
        PolicierParcel,
        PreferredColorGroupParcel,
        SLAClassParcel,
        TlocParcel,
        StandardCommunityParcel,
        LocalDomainParcel,
        FQDNDomainParcel,
        IPSSignatureParcel,
        URLAllowParcel,
        URLBlockParcel,
        SecurityPortParcel,
        ProtocolListParcel,
        GeoLocationListParcel,
        SecurityZoneListParcel,
        SecurityApplicationListParcel,
        SecurityDataPrefixParcel,
    ],
    Field(discriminator="type"),
]

POLICY_OBJECT_PAYLOAD_ENDPOINT_MAPPING: Mapping[type, str] = {
    AppProbeParcel: "app-probe",
    ApplicationListParcel: "app-list",
    ColorParcel: "color",
    DataPrefixParcel: "data-prefix",
    ExpandedCommunityParcel: "expanded-community",
    FowardingClassParcel: "class",
    IPv6DataPrefixParcel: "data-ipv6-prefix",
    IPv6PrefixListParcel: "ipv6-prefix",
    PrefixListParcel: "prefix",
    PolicierParcel: "policer",
    PreferredColorGroupParcel: "preferred-color-group",
    SLAClassParcel: "sla-class",
    TlocParcel: "tloc",
    StandardCommunityParcel: "standard-community",
    LocalDomainParcel: "security-localdomain",
    FQDNDomainParcel: "security-fqdn",
    IPSSignatureParcel: "security-ipssignature",
    URLAllowParcel: "security-urllist",
    URLBlockParcel: "security-urllist",
    SecurityPortParcel: "security-port",
    ProtocolListParcel: "security-protocolname",
    GeoLocationListParcel: "security-geolocation",
    SecurityZoneListParcel: "security-zone",
    SecurityApplicationListParcel: "security-localapp",
    SecurityDataPrefixParcel: "security-data-ip-prefix",
}

__all__ = (
    "AnyPolicyObjectParcel",
    "ApplicationFamilyListEntry",
    "ApplicationListEntry",
    "ApplicationListParcel",
    "AppProbeEntry",
    "AppProbeMapItem",
    "AppProbeParcel",
    "BaseURLListEntry",
    "ColorEntry",
    "ColorParcel",
    "DataPrefixEntry",
    "DataPrefixParcel",
    "ExpandedCommunityParcel",
    "FallbackBestTunnel",
    "FowardingClassParcel",
    "FowardingClassQueueEntry",
    "FQDNDomainParcel",
    "FQDNListEntry",
    "GeoLocationListEntry",
    "GeoLocationListParcel",
    "IPSSignatureListEntry",
    "IPSSignatureParcel",
    "IPv6DataPrefixEntry",
    "IPv6DataPrefixParcel",
    "IPv6PrefixListEntry",
    "IPv6PrefixListParcel",
    "LocalDomainListEntry",
    "LocalDomainParcel",
    "PolicierEntry",
    "PolicierParcel",
    "Preference",
    "PreferredColorGroupEntry",
    "PreferredColorGroupParcel",
    "PrefixListEntry",
    "PrefixListParcel",
    "ProtocolListEntry",
    "ProtocolListParcel",
    "SecurityApplicationFamilyListEntry",
    "SecurityApplicationListEntry",
    "SecurityApplicationListParcel",
    "SecurityDataPrefixEntry",
    "SecurityDataPrefixParcel",
    "SecurityPortListEntry",
    "SecurityPortParcel",
    "SecurityZoneListEntry",
    "SecurityZoneListParcel",
    "SLAAppProbeClass",
    "SLAClassCriteria",
    "SLAClassListEntry",
    "SLAClassParcel",
    "StandardCommunityEntry",
    "StandardCommunityParcel",
    "TlocEntry",
    "TlocParcel",
    "URLAllowParcel",
    "URLBlockParcel",
)


def __dir__() -> "List[str]":
    return list(__all__)



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from typing import List

from pydantic import AliasPath, BaseModel, ConfigDict, Field

from catalystwan.api.configuration_groups.parcel import Global, _ParcelBase, as_global


class BaseURLListEntry(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    pattern: Global[str]


class BaseURLParcel(_ParcelBase):
    entries: List[BaseURLListEntry] = Field(default=[], validation_alias=AliasPath("data", "entries"))

    def add_url(self, pattern: str):
        self.entries.append(BaseURLListEntry(pattern=as_global(pattern)))


class URLAllowParcel(BaseURLParcel):
    parcel_type: str = Field(default="urlallowed", validation_alias="type", serialization_alias="type")


class URLBlockParcel(BaseURLParcel):
    parcel_type: str = Field(default="urlblocked", validation_alias="type", serialization_alias="type")



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from catalystwan.session import ManagerSession

from catalystwan.endpoints.configuration_group import (
    ConfigGroupAssociatePayload,
    ConfigGroupCreationPayload,
    ConfigGroupCreationResponse,
    ConfigGroupDeployPayload,
    ConfigGroupDeployResponse,
    ConfigGroupDisassociateResponse,
    ConfigGroupEditPayload,
    ConfigGroupEditResponse,
    ConfigGroupResponsePayload,
    ConfigGroupVariablesCreatePayload,
    ConfigGroupVariablesCreateResponse,
    ConfigGroupVariablesEditPayload,
    ConfigurationGroup,
    DeviceId,
    ProfileId,
    Solution,
)


class ConfigGroupAPI:
    def __init__(self, session: ManagerSession):
        self.session = session
        self.endpoint = ConfigurationGroup(session)

    def associate(self, cg_id: str, device_ids: list) -> None:
        """
        Associates given config-group to specified list of devices
        """
        devices = []

        for device_id in device_ids:
            devices.append(DeviceId(id=device_id))

        payload = ConfigGroupAssociatePayload(devices=devices)

        self.endpoint.associate(config_group_id=cg_id, payload=payload)

    def create(self, name: str, description: str, solution: Solution, profile_ids: list) -> ConfigGroupCreationResponse:
        """
        Creates new config-group
        """
        profiles = []

        for profile_id in profile_ids:
            profiles.append(ProfileId(id=profile_id))
        cg_payload = ConfigGroupCreationPayload(
            name=name, description=description, solution=solution, profiles=profiles
        )

        return self.endpoint.create_config_group(cg_payload)

    def create_variables(
        self, cg_id: str, device_ids: list, suggestions: bool = True
    ) -> ConfigGroupVariablesCreateResponse:
        """
        Creates device specific variable data in given config-group
        """
        payload = ConfigGroupVariablesCreatePayload(deviceIds=device_ids, suggestions=suggestions)
        return self.endpoint.create_variables(config_group_id=cg_id, payload=payload)

    def delete(self, cg_id: str) -> None:
        """
        Deletes existing config-group with given ID
        """
        self.endpoint.delete_config_group(cg_id)

    def deploy(self, cg_id: str, device_ids: list) -> ConfigGroupDeployResponse:
        """
        Deploys specified config-group config to given list of devices
        """
        devices = []
        for device_id in device_ids:
            devices.append(DeviceId(id=device_id))

        payload = ConfigGroupDeployPayload(devices=devices)
        return self.endpoint.deploy(config_group_id=cg_id, payload=payload)

    def disassociate(self, cg_id: str, device_ids: list) -> ConfigGroupDisassociateResponse:
        """
        Disassociates given list of devices from the specified config-group
        """
        devices = []
        for device_id in device_ids:
            devices.append(DeviceId(id=device_id))

        payload = ConfigGroupAssociatePayload(devices=devices)
        return self.endpoint.disassociate(config_group_id=cg_id, payload=payload)

    def edit(
        self, cg_id: str, name: str, description: str, solution: Solution, profile_ids: list
    ) -> ConfigGroupEditResponse:
        """
        Modifies feature profiles in existing config-group
        """
        profiles = []

        for profile_id in profile_ids:
            profiles.append(ProfileId(id=profile_id))
        payload = ConfigGroupEditPayload(name=name, description=description, solution=solution, profiles=profiles)

        return self.endpoint.edit_config_group(config_group_id=cg_id, payload=payload)

    def get(self) -> ConfigGroupResponsePayload:
        """
        Gets list of existing config-groups
        """
        return self.endpoint.get()

    def update_variables(self, cg_id: str, solution: Solution, device_variables: list) -> None:
        """
        Updates device specific variable data in given config-group
        """
        payload = ConfigGroupVariablesEditPayload(solution=solution, devices=device_variables)

        self.endpoint.update_variables(config_group_id=cg_id, payload=payload)



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from __future__ import annotations

from enum import Enum
from pathlib import Path
from typing import ClassVar, List, Optional

from attr import define, field  # type: ignore
from pydantic import ConfigDict

from catalystwan.api.templates.feature_template import FeatureTemplate
from catalystwan.dataclasses import User


class AuthenticationOrder(Enum):
    LOCAL = "local"
    RADIUS = "radius"
    TACACS = "tacacs"


class TacacsAuthenticationMethod(Enum):
    PAP = "pap"


class Action(Enum):
    ACCEPT = "accept"
    DENY = "deny"


class VpnType(Enum):
    VPN_TRANSPORT = 0
    VPN_MANAGMENT = 512


@define
class TacacsServer:
    """Default values from documentations."""

    address: str
    auth_port: int = field(default=49)
    secret_key: Optional[str] = field(default=None)
    source_interface: Optional[str] = field(default=None)
    vpn: int = field(default=0)
    priority: int = field(default=0)


@define
class RadiusServer:
    """Default values from documentations."""

    address: str
    secret_key: Optional[str] = field(default=None)
    source_interface: Optional[str] = field(default=None)
    acct_port: int = field(default=1813)
    auth_port: int = field(default=1812)
    tag: Optional[str] = field(default=None)
    timeout: int = field(default=5)
    vpn: int = field(default=0)
    priority: int = field(default=0)


@define
class AuthTask:
    name: str
    default_action: Action = field(default=Action.ACCEPT)


class AAAModel(FeatureTemplate):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    payload_path: ClassVar[Path] = Path(__file__).parent / "feature" / "aaa.json.j2"
    type: ClassVar[str] = "aaa"  # AAA

    auth_order: List[AuthenticationOrder]
    auth_fallback: bool
    auth_disable_audit_logs: bool
    auth_admin_order: bool
    auth_disable_netconf_logs: bool
    auth_radius_servers: List[str] = []

    local_users: List[User] = []

    accounting: bool = True

    tacacs_authentication: TacacsAuthenticationMethod = TacacsAuthenticationMethod.PAP
    tacacs_timeout: int = 5
    tacacs_servers: List[TacacsServer] = []
    radius_retransmit: int = 3
    radius_timeout: int = 5
    radius_servers: List[RadiusServer] = []



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from typing import List, Literal, Optional, Union
from uuid import UUID

from pydantic import BaseModel, ConfigDict, Field

from catalystwan.api.configuration_groups.parcel import Default, Global, Variable
from catalystwan.models.configuration.feature_profile.common import Prefix


class AggregatePrefix(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    prefix: Prefix
    as_set: Optional[Union[Global[bool], Variable, Default[bool]]] = Field(
        serialization_alias="asSet", validation_alias="asSet", default=Default[bool](value=False)
    )
    summary_only: Optional[Union[Global[bool], Variable, Default[bool]]] = Field(
        serialization_alias="summaryOnly", validation_alias="summaryOnly", default=Default[bool](value=False)
    )


class AggregatePrefixIPv6(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    prefix: Union[Global[str], Variable]
    as_set: Optional[Union[Global[bool], Variable, Default[bool]]] = Field(
        serialization_alias="asSet", validation_alias="asSet", default=Default[bool](value=False)
    )
    summary_only: Optional[Union[Global[bool], Variable, Default[bool]]] = Field(
        serialization_alias="summaryOnly", validation_alias="summaryOnly", default=Default[bool](value=False)
    )


class NetworkPrefix(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    prefix: Prefix


class NetworkPrefixIPv6(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    prefix: Union[Global[str], Variable]


RedistributeProtocol = Literal[
    "static",
    "connected",
    "omp",
    "nat",
    "ospf",
    "ospfv3",
    "eigrp",
]


RedistributeProtocolIPv6 = Literal[
    "static",
    "connected",
    "omp",
    "ospf",
]


class RedistributedRoute(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    protocol: Union[Global[RedistributeProtocol], Variable]
    route_policy: Optional[Union[Default[None], Global[UUID]]] = Field(
        serialization_alias="routePolicy", validation_alias="routePolicy"
    )


class RedistributedRouteIPv6(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    protocol: Union[Global[RedistributeProtocolIPv6], Variable]
    route_policy: Optional[Union[Default[None], Global[UUID]]] = Field(
        serialization_alias="routePolicy", validation_alias="routePolicy"
    )


class AddressFamilyIPv4(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    aggregate_address: Optional[List[AggregatePrefix]] = Field(
        serialization_alias="aggregateAddress", validation_alias="aggregateAddress"
    )
    network: Optional[List[NetworkPrefix]] = None
    paths: Optional[Union[Global[int], Variable, Default[None]]] = None
    originate: Optional[Union[Global[bool], Variable, Default[bool]]] = Default[bool](value=False)
    name: Optional[Union[Default[None], Global[UUID]]] = None
    filter: Optional[Union[Global[bool], Variable, Default[bool]]] = Default[bool](value=False)
    redistribute: Optional[List[RedistributedRoute]] = None


class PolicyType(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    policy_type: Global[str] = Field(serialization_alias="policyType", validation_alias="policyType")


class PolicyTypeWithThreshold(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    policy_type: Global[str] = Field(serialization_alias="policyType", validation_alias="policyType")
    prefix_number: Union[Global[int], Variable] = Field(serialization_alias="prefixNum", validation_alias="prefixNum")
    threshold: Union[Global[int], Variable, Default[int]] = Default[int](value=75)


class PolicyTypeWithRestart(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    policy_type: Global[str] = Field(serialization_alias="policyType", validation_alias="policyType")
    prefix_number: Union[Global[int], Variable] = Field(serialization_alias="prefixNum", validation_alias="prefixNum")
    threshold: Union[Global[int], Variable, Default[int]] = Default[int](value=75)
    restart_interval: Union[Global[int], Variable]


class AddressFamilyIPv6(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    aggregate_address: Optional[List[AggregatePrefixIPv6]] = Field(
        serialization_alias="ipv6AggregateAddress", validation_alias="ipv6AggregateAddress"
    )
    network: Optional[List[NetworkPrefixIPv6]] = Field(
        serialization_alias="ipv6Network", validation_alias="ipv6Network"
    )
    paths: Optional[Union[Global[int], Variable, Default[None]]] = None
    originate: Optional[Union[Global[bool], Variable, Default[bool]]] = Default[bool](value=False)
    name: Optional[Union[Default[None], Global[UUID]]] = None
    filter: Optional[Union[Global[bool], Variable, Default[bool]]] = Default[bool](value=False)
    redistribute: Optional[List[RedistributedRouteIPv6]] = None


class BgpAddressFamily(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    family_type: Global[str] = Field(serialization_alias="familyType", validation_alias="familyType")
    max_prefix_config: Optional[Union[PolicyType, PolicyTypeWithRestart, PolicyTypeWithThreshold]] = Field(
        serialization_alias="maxPrefixConfig", validation_alias="maxPrefixConfig"
    )
    in_route_policy: Optional[Union[Default[None], Global[UUID]]] = Field(
        serialization_alias="inRoutePolicy", validation_alias="inRoutePolicy"
    )
    out_route_policy: Optional[Union[Default[None], Global[UUID]]] = Field(
        serialization_alias="outRoutePolicy", validation_alias="outRoutePolicy"
    )


class BgpIPv4Neighbor(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    address: Union[Global[str], Variable]
    description: Optional[Union[Global[str], Variable, Default[None]]] = None
    shutdown: Optional[Union[Global[bool], Variable, Default[bool]]] = Default[bool](value=False)
    remote_as: Union[Global[int], Variable] = Field(serialization_alias="remoteAs", validation_alias="remoteAs")
    local_as: Union[Global[int], Variable] = Field(
        serialization_alias="localAs", validation_alias="localAs", default=None
    )
    keepalive: Optional[Union[Global[int], Variable, Default[int]]] = Default[int](value=60)
    holdtime: Optional[Union[Global[int], Variable, Default[int]]] = Default[int](value=180)
    interface_name: Optional[Union[Global[str], Variable, Default[None]]] = Field(
        serialization_alias="ifName", validation_alias="ifName", default=None
    )
    next_hop_self: Optional[Union[Global[bool], Variable, Default[bool]]] = Field(
        serialization_alias="nextHopSelf", validation_alias="nextHopSelf", default=Default[bool](value=False)
    )
    send_community: Optional[Union[Global[bool], Variable, Default[bool]]] = Field(
        serialization_alias="sendCommunity", validation_alias="sendCommunity", default=Default[bool](value=True)
    )
    send_ext_community: Optional[Union[Global[bool], Variable, Default[bool]]] = Field(
        serialization_alias="sendExtCommunity", validation_alias="sendExtCommunity", default=Default[bool](value=True)
    )
    ebgp_multihop: Optional[Union[Global[int], Variable, Default[int]]] = Field(
        serialization_alias="ebgpMultihop", validation_alias="ebgpMultihop", default=Default[int](value=1)
    )
    password: Optional[Union[Global[str], Variable, Default[None]]] = None
    send_label: Optional[Union[Global[bool], Variable, Default[bool]]] = Field(
        serialization_alias="sendLabel", validation_alias="sendLabel", default=Default[bool](value=False)
    )
    as_override: Optional[Union[Global[bool], Variable, Default[bool]]] = Field(
        serialization_alias="asOverride", validation_alias="asOverride", default=Default[bool](value=False)
    )
    as_number: Optional[Union[Global[int], Variable, Default[None]]] = Field(
        serialization_alias="asNumber", validation_alias="asNumber", default=None
    )
    address_family: Optional[BgpAddressFamily] = Field(
        serialization_alias="addressFamily", validation_alias="addressFamily"
    )


class BgpIPv6Neighbor(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    address: Union[Global[str], Variable]
    description: Optional[Union[Global[str], Variable, Default[None]]] = None
    shutdown: Optional[Union[Global[bool], Variable, Default[bool]]] = Default[bool](value=False)
    remote_as: Union[Global[int], Variable] = Field(serialization_alias="remoteAs", validation_alias="remoteAs")
    local_as: Union[Global[int], Variable] = Field(
        serialization_alias="localAs", validation_alias="localAs", default=None
    )
    keepalive: Optional[Union[Global[int], Variable, Default[int]]] = Default[int](value=60)
    holdtime: Optional[Union[Global[int], Variable, Default[int]]] = Default[int](value=180)
    interface_name: Optional[Union[Global[str], Variable, Default[None]]] = Field(
        serialization_alias="ifName", validation_alias="ifName", default=None
    )
    next_hop_self: Optional[Union[Global[bool], Variable, Default[bool]]] = Field(
        serialization_alias="nextHopSelf", validation_alias="nextHopSelf", default=Default[bool](value=False)
    )
    send_community: Optional[Union[Global[bool], Variable, Default[bool]]] = Field(
        serialization_alias="sendCommunity", validation_alias="sendCommunity", default=Default[bool](value=True)
    )
    send_ext_community: Optional[Union[Global[bool], Variable, Default[bool]]] = Field(
        serialization_alias="sendExtCommunity", validation_alias="sendExtCommunity", default=Default[bool](value=True)
    )
    ebgp_multihop: Optional[Union[Global[int], Variable, Default[int]]] = Field(
        serialization_alias="ebgpMultihop", validation_alias="ebgpMultihop", default=Default[int](value=1)
    )
    password: Optional[Union[Global[str], Variable, Default[None]]] = None
    send_label: Optional[Union[Global[bool], Variable, Default[bool]]] = Field(
        serialization_alias="sendLabel", validation_alias="sendLabel", default=Default[bool](value=False)
    )
    as_override: Optional[Union[Global[bool], Variable, Default[bool]]] = Field(
        serialization_alias="asOverride", validation_alias="asOverride", default=Default[bool](value=False)
    )
    as_number: Optional[Union[Global[int], Variable, Default[None]]] = Field(
        serialization_alias="asNumber", validation_alias="asNumber", default=None
    )
    address_family: Optional[BgpAddressFamily] = Field(
        serialization_alias="addressFamily", validation_alias="addressFamily"
    )


class BgpData(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    as_num: Union[Global[int], Variable] = Field(serialization_alias="asNum", validation_alias="asNum")
    router_id: Optional[Union[Global[str], Variable, Default[None]]] = Field(
        serialization_alias="routerId", validation_alias="routerId", default=None
    )
    propagate_aspath: Optional[Union[Global[bool], Variable, Default[bool]]] = Field(
        serialization_alias="propagateAspath", validation_alias="propagateAspath", default=Default[bool](value=False)
    )
    propagate_community: Optional[Union[Global[bool], Variable, Default[bool]]] = Field(
        serialization_alias="propagateCommunity",
        validation_alias="propagateCommunity",
        default=Default[bool](value=False),
    )
    external: Optional[Union[Global[int], Variable, Default[int]]] = Default[int](value=20)
    internal: Optional[Union[Global[int], Variable, Default[int]]] = Default[int](value=200)
    local: Optional[Union[Global[int], Variable, Default[int]]] = Default[int](value=20)
    keepalive: Optional[Union[Global[int], Variable, Default[int]]] = Default[int](value=60)
    holdtime: Optional[Union[Global[int], Variable, Default[int]]] = Default[int](value=180)
    always_compare: Optional[Union[Global[bool], Variable, Default[bool]]] = Field(
        serialization_alias="alwaysCompare", validation_alias="alwaysCompare", default=Default[bool](value=False)
    )
    deterministic: Optional[Union[Global[bool], Variable, Default[bool]]] = Default[bool](value=False)
    missing_as_worst: Optional[Union[Global[bool], Variable, Default[bool]]] = Field(
        serialization_alias="missingAsWorst", validation_alias="missingAsWorst", default=Default[bool](value=False)
    )
    compare_router_id: Optional[Union[Global[bool], Variable, Default[bool]]] = Field(
        serialization_alias="compareRouterId", validation_alias="compareRouterId", default=Default[bool](value=False)
    )
    multipath_relax: Optional[Union[Global[bool], Variable, Default[bool]]] = Field(
        serialization_alias="multipathRelax", validation_alias="multipathRelax", default=Default[bool](value=False)
    )
    neighbor: Optional[List[BgpIPv4Neighbor]] = None
    ipv6_neighbor: Optional[List[BgpIPv6Neighbor]] = Field(
        serialization_alias="ipv6Neighbor", validation_alias="ipv6Neighbor", default=None
    )
    address_family: Optional[AddressFamilyIPv4] = Field(
        serialization_alias="addressFamily", validation_alias="addressFamily", default=None
    )
    ipv6_address_family: Optional[AddressFamilyIPv6] = Field(
        serialization_alias="ipv6AddressFamily", validation_alias="ipv6AddressFamily", default=None
    )


class BgpCreationPayload(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    name: str
    description: Optional[str] = None
    data: BgpData
    metadata: Optional[dict] = None



# Copyright 2023 Cisco Systems, Inc. and its affiliates

import datetime
from functools import wraps
from ipaddress import IPv4Address, IPv4Network, IPv6Network
from typing import Any, Dict, List, MutableSequence, Optional, Protocol, Sequence, Set, Tuple, Union
from uuid import UUID

from pydantic import BaseModel, ConfigDict, Field, RootModel, model_validator
from typing_extensions import Annotated, Literal

from catalystwan.models.common import ServiceChainNumber, TLOCColor, check_fields_exclusive
from catalystwan.models.misc.application_protocols import ApplicationProtocol
from catalystwan.models.policy.lists_entries import EncapType
from catalystwan.typed_list import DataSequence


def port_set_and_ranges_to_str(ports: Set[int] = set(), port_ranges: List[Tuple[int, int]] = []) -> str:
    assert ports or port_ranges
    ports_str = " ".join(f"{port_begin}-{port_end}" for port_begin, port_end in port_ranges)
    ports_str += " " if ports_str else ""
    ports_str += " ".join(str(p) for p in ports)
    return ports_str


def networks_to_str(networks: Sequence[Union[IPv4Network, IPv6Network]]) -> str:
    return " ".join(str(net) for net in networks)


PLPEntryType = Literal[
    "low",
    "high",
]

DNSEntryType = Literal[
    "request",
    "response",
]

TrafficTargetType = Literal[
    "access",
    "core",
    "service",
]

DestinationRegion = Literal[
    "primary-region",
    "secondary-region",
    "other-region",
]

OriginProtocol = Literal[
    "aggregate",
    "bgp",
    "bgp-external",
    "bgp-internal",
    "connected",
    "eigrp",
    "ospf",
    "ospf-inter-area",
    "ospf-intra-area",
    "ospf-external1",
    "ospf-external2",
    "rip",
    "static",
    "eigrp-summary",
    "eigrp-internal",
    "eigrp-external",
    "lisp",
    "nat-dia",
    "natpool",
    "isis",
    "isis-level1",
    "isis-level2",
]

PathType = Literal[
    "hierarchical-path",
    "direct-path",
    "transport-gateway-path",
]

SequenceIpType = Literal[
    "ipv4",
    "ipv6",
    "all",
]

PolicyActionType = Literal[
    "drop",
    "accept",
    "pass",
    "inspect",
    "reject",
]

SequenceType = Literal[
    "applicationFirewall",
    "data",
    "serviceChaining",
    "trafficEngineering",
    "qos",
    "zoneBasedFW",
    "tloc",
    "route",
    "acl",
    "aclv6",
    "deviceaccesspolicy",
    "deviceaccesspolicyv6",
]


Optimized = Literal[
    "true",
    "false",
]

DNSTypeEntryType = Literal[
    "host",
    "umbrella",
]

LossProtectionType = Literal[
    "fecAdaptive",
    "fecAlways",
    "packetDuplication",
]

MultiRegionRole = Literal[
    "border-router",
    "edge-router",
]

ServiceType = Literal[
    "FW",
    "IDP",
    "IDS",
    "netsvc1",
    "netsvc2",
    "netsvc3",
    "netsvc4",
]

TLOCActionType = Literal[
    "strict",
    "primary",
    "backup",
    "ecmp",
]

Carrier = Literal[
    "default",
    "carrier1",
    "carrier2",
    "carrier3",
    "carrier4",
    "carrier5",
    "carrier6",
    "carrier7",
    "carrier8",
]

DeviceAccessProtocol = Literal[22, 161]


class Reference(BaseModel):
    ref: UUID


class VariableName(BaseModel):
    vip_variable_name: str = Field(serialization_alias="vipVariableName", validation_alias="vipVariableName")


class LocalTLOCListEntryValue(BaseModel):
    color: TLOCColor
    encap: EncapType
    restrict: Optional[str] = None


class TLOCEntryValue(BaseModel):
    ip: IPv4Address
    color: TLOCColor
    encap: EncapType


class ServiceChainEntryValue(BaseModel):
    type: ServiceChainNumber = Field(default="SC1")
    vpn: str
    restrict: Optional[str] = None
    local: Optional[str] = None
    tloc: Optional[TLOCEntryValue] = None


class PacketLengthEntry(BaseModel):
    field: Literal["packetLength"] = "packetLength"
    value: str = Field(description="0-65536 range or single number")

    @staticmethod
    def from_range(packet_lengths: Tuple[int, int]) -> "PacketLengthEntry":
        if packet_lengths[0] == packet_lengths[1]:
            return PacketLengthEntry(value=str(packet_lengths[0]))
        return PacketLengthEntry(value=f"{packet_lengths[0]}-{packet_lengths[1]}")


class PLPEntry(BaseModel):
    field: Literal["plp"] = "plp"
    value: PLPEntryType


class ProtocolEntry(BaseModel):
    field: Literal["protocol"] = "protocol"
    value: str = Field(description="0-255 single numbers separate by space")
    app: Optional[str] = None

    @staticmethod
    def from_protocol_set(protocols: Set[int]) -> "ProtocolEntry":
        return ProtocolEntry(value=" ".join(str(p) for p in protocols))

    @staticmethod
    def from_application_protocols(app_prots: List[ApplicationProtocol]) -> "ProtocolEntry":
        return ProtocolEntry(
            value=" ".join(p.protocol_as_string_of_numbers() for p in app_prots),
            app=" ".join(p.name for p in app_prots),
        )


class DSCPEntry(BaseModel):
    field: Literal["dscp"] = "dscp"
    value: str = Field(description="0-63 single numbers separate by space")


class SourceIPEntry(BaseModel):
    field: Literal["sourceIp"] = "sourceIp"
    value: str = Field(description="IP network specifiers separate by space")

    @staticmethod
    def from_ipv4_networks(networks: List[IPv4Network]) -> "SourceIPEntry":
        return SourceIPEntry(value=networks_to_str(networks))


class SourceIPv6Entry(BaseModel):
    field: Literal["sourceIpv6"] = "sourceIpv6"
    value: str = Field(description="IPv6 network specifiers separate by space")

    @staticmethod
    def from_ipv6_networks(networks: List[IPv6Network]) -> "SourceIPv6Entry":
        return SourceIPv6Entry(value=networks_to_str(networks))


class IPAddressEntry(BaseModel):
    field: Literal["ipAddress"] = "ipAddress"
    value: IPv4Address


class SourcePortEntry(BaseModel):
    field: Literal["sourcePort"] = "sourcePort"
    value: str = Field(description="0-65535 range or separate by space")

    @staticmethod
    def from_port_set_and_ranges(ports: Set[int] = set(), port_ranges: List[Tuple[int, int]] = []) -> "SourcePortEntry":
        return SourcePortEntry(value=port_set_and_ranges_to_str(ports, port_ranges))


class DestinationIPEntry(BaseModel):
    field: Literal["destinationIp"] = "destinationIp"
    value: str

    @staticmethod
    def from_ipv4_networks(networks: List[IPv4Network]) -> "DestinationIPEntry":
        return DestinationIPEntry(value=networks_to_str(networks))


class DestinationIPv6Entry(BaseModel):
    field: Literal["destinationIpv6"] = "destinationIpv6"
    value: str

    @staticmethod
    def from_ipv6_networks(networks: List[IPv6Network]) -> "DestinationIPv6Entry":
        return DestinationIPv6Entry(value=networks_to_str(networks))


class DestinationPortEntry(BaseModel):
    field: Literal["destinationPort"] = "destinationPort"
    value: str = Field(description="0-65535 range or separate by space")
    app: Optional[str] = None

    @staticmethod
    def from_port_set_and_ranges(
        ports: Set[int] = set(), port_ranges: List[Tuple[int, int]] = []
    ) -> "DestinationPortEntry":
        return DestinationPortEntry(value=port_set_and_ranges_to_str(ports, port_ranges))

    @staticmethod
    def from_application_protocols(app_prots: List[ApplicationProtocol]) -> "DestinationPortEntry":
        return DestinationPortEntry(
            value=" ".join(p.port for p in app_prots if p.port),
            app=" ".join(p.name for p in app_prots),
        )


class TCPEntry(BaseModel):
    field: Literal["tcp"] = "tcp"
    value: Literal["syn"] = "syn"


class DNSEntry(BaseModel):
    field: Literal["dns"] = "dns"
    value: DNSEntryType


class TrafficToEntry(BaseModel):
    field: Literal["trafficTo"] = "trafficTo"
    value: TrafficTargetType


class DestinationRegionEntry(BaseModel):
    field: Literal["destinationRegion"] = "destinationRegion"
    value: DestinationRegion


class SourceFQDNEntry(BaseModel):
    field: Literal["sourceFqdn"] = "sourceFqdn"
    value: str = Field(max_length=120)


class DestinationFQDNEntry(BaseModel):
    field: Literal["destinationFqdn"] = "destinationFqdn"
    value: str = Field(max_length=120)


class SourceGeoLocationEntry(BaseModel):
    field: Literal["sourceGeoLocation"] = "sourceGeoLocation"
    value: str = Field(description="Space separated list of ISO3166 country codes")


class DestinationGeoLocationEntry(BaseModel):
    field: Literal["destinationGeoLocation"] = "destinationGeoLocation"
    value: str = Field(description="Space separated list of ISO3166 country codes")


class ProtocolNameEntry(BaseModel):
    field: Literal["protocolName"] = "protocolName"
    value: str

    @staticmethod
    def from_application_protocols(app_prots: List[ApplicationProtocol]) -> "ProtocolNameEntry":
        return ProtocolNameEntry(value=" ".join(p.name for p in app_prots))


class ForwardingClassEntry(BaseModel):
    field: Literal["forwardingClass"] = "forwardingClass"
    value: str = Field(max_length=32)


class NATPoolEntry(BaseModel):
    field: Literal["pool"] = "pool"
    value: str


class UseVPNEntry(BaseModel):
    field: Literal["useVpn"] = "useVpn"
    value: str = "0"


class FallBackEntry(BaseModel):
    field: Literal["fallback"] = "fallback"
    value: str = ""


class NextHopEntry(BaseModel):
    field: Literal["nextHop"] = "nextHop"
    value: IPv4Address


class NextHopLooseEntry(BaseModel):
    field: Literal["nextHopLoose"] = "nextHopLoose"
    value: str


class OMPTagEntry(BaseModel):
    field: Literal["ompTag"] = "ompTag"
    value: str = Field(description="Number in range 0-4294967295")


class OriginEntry(BaseModel):
    field: Literal["origin"] = "origin"
    value: OriginProtocol


class OriginatorEntry(BaseModel):
    field: Literal["originator"] = "originator"
    value: IPv4Address


class PreferenceEntry(BaseModel):
    field: Literal["preference"] = "preference"
    value: str = Field(description="Number in range 0-4294967295")


class PathTypeEntry(BaseModel):
    field: Literal["pathType"] = "pathType"
    value: PathType


class RegionEntry(BaseModel):
    field: Literal["regionId"] = "regionId"
    value: str


class RoleEntry(BaseModel):
    field: Literal["role"] = "role"
    value: MultiRegionRole


class SiteEntry(BaseModel):
    field: Literal["siteId"] = "siteId"
    value: str = Field(description="Site ID numeric value")


class LocalTLOCListEntry(BaseModel):
    field: Literal["localTlocList"] = "localTlocList"
    value: LocalTLOCListEntryValue


class DNSTypeEntry(BaseModel):
    field: Literal["dnsType"] = "dnsType"
    value: DNSTypeEntryType


class ServiceChainEntry(BaseModel):
    field: Literal["serviceChain"] = "serviceChain"
    value: ServiceChainEntryValue


class VPNEntry(BaseModel):
    field: Literal["vpn"] = "vpn"
    value: str


class TLOCEntry(BaseModel):
    field: Literal["tloc"] = "tloc"
    value: TLOCEntryValue


class CommunityEntry(BaseModel):
    field: Literal["community"] = "community"
    value: str = Field(description="Example: 1000:10000 or internet or local-AS or no advertise or no-export")


class CommunityAdditiveEntry(BaseModel):
    field: Literal["communityAdditive"] = "communityAdditive"
    value: Literal["true"] = "true"


class CarrierEntry(BaseModel):
    field: Literal["carrier"] = "carrier"
    value: Carrier


class DomainIDEntry(BaseModel):
    field: Literal["domainId"] = "domainId"
    value: str = Field(description="Number in range 1-4294967295")


class GroupIDEntry(BaseModel):
    field: Literal["groupId"] = "groupId"
    value: str = Field(description="Number in range 0-4294967295")


class NextHeaderEntry(BaseModel):
    field: Literal["nextHeader"] = "nextHeader"
    value: str = Field(description="0-63 single numbers separate by space")


class TrafficClassEntry(BaseModel):
    field: Literal["trafficClass"] = "trafficClass"
    value: str = Field(description="Number in range 0-63")


class NATVPNEntry(RootModel):
    root: List[Union[UseVPNEntry, FallBackEntry]]

    @staticmethod
    def from_nat_vpn(fallback: bool, vpn: int = 0) -> "NATVPNEntry":
        if fallback:
            return NATVPNEntry(root=[UseVPNEntry(value=str(vpn)), FallBackEntry()])
        return NATVPNEntry(root=[UseVPNEntry(value=str(vpn))])


class SourceDataPrefixListEntry(BaseModel):
    field: Literal["sourceDataPrefixList"] = "sourceDataPrefixList"
    ref: UUID


class SourceDataIPv6PrefixListEntry(BaseModel):
    field: Literal["sourceDataIpv6PrefixList"] = "sourceDataIpv6PrefixList"
    ref: UUID


class DestinationDataPrefixListEntry(BaseModel):
    field: Literal["destinationDataPrefixList"] = "destinationDataPrefixList"
    ref: UUID


class DestinationDataIPv6PrefixListEntry(BaseModel):
    field: Literal["destinationDataIpv6PrefixList"] = "destinationDataIpv6PrefixList"
    ref: UUID


class DNSAppListEntry(BaseModel):
    field: Literal["dnsAppList"] = "dnsAppList"
    ref: UUID


class AppListEntry(BaseModel):
    field: Literal["appList"] = "appList"
    ref: UUID


class SourceFQDNListEntry(BaseModel):
    field: Literal["sourceFqdnList"] = "sourceFqdnList"
    ref: UUID


class DestinationFQDNListEntry(BaseModel):
    field: Literal["destinationFqdnList"] = "destinationFqdnList"
    ref: UUID


class SourceGeoLocationListEntry(BaseModel):
    field: Literal["sourceGeoLocationList"] = "sourceGeoLocationList"
    ref: UUID


class DestinationGeoLocationListEntry(BaseModel):
    field: Literal["destinationGeoLocationList"] = "destinationGeoLocationList"
    ref: UUID


class ProtocolNameListEntry(BaseModel):
    field: Literal["protocolNameList"] = "protocolNameList"
    ref: UUID


class SourcePortListEntry(BaseModel):
    field: Literal["sourcePortList"] = "sourcePortList"
    ref: UUID


class DestinationPortListEntry(BaseModel):
    field: Literal["destinationPortList"] = "destinationPortList"
    ref: UUID


class RuleSetListEntry(BaseModel):
    field: Literal["ruleSetList"] = "ruleSetList"
    ref: str

    @staticmethod
    def from_rule_set_ids(rule_set_ids: Set[UUID]) -> "RuleSetListEntry":
        return RuleSetListEntry(ref=" ".join(str(rule_set_ids)))


class PolicerListEntry(BaseModel):
    field: Literal["policer"] = "policer"
    ref: UUID


class TLOCListEntry(BaseModel):
    field: Literal["tlocList"] = "tlocList"
    ref: UUID


class PrefferedColorGroupListEntry(BaseModel):
    field: Literal["preferredColorGroup"] = "preferredColorGroup"
    ref: UUID
    color_restrict: bool = Field(False, serialization_alias="colorRestrict", validation_alias="colorRestrict")
    model_config = ConfigDict(populate_by_name=True)


class ColorListEntry(BaseModel):
    field: Literal["colorList"] = "colorList"
    ref: UUID


class CommunityListEntry(BaseModel):
    field: Literal["community"] = "community"
    ref: UUID


class ExpandedCommunityListEntry(BaseModel):
    field: Literal["expandedCommunity"] = "expandedCommunity"
    ref: UUID


class SiteListEntry(BaseModel):
    field: Literal["siteList"] = "siteList"
    ref: UUID


class VPNListEntry(BaseModel):
    field: Literal["vpnList"] = "vpnList"
    ref: UUID


class PrefixListEntry(BaseModel):
    field: Literal["prefixList"] = "prefixList"
    ref: UUID


class RegionListEntry(BaseModel):
    field: Literal["regionList"] = "regionList"
    ref: UUID


class ClassMapListEntry(BaseModel):
    field: Literal["class"] = "class"
    ref: UUID


class ServiceEntryValue(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    type: ServiceType
    vpn: str
    tloc: Optional[TLOCEntryValue] = None
    tloc_list: Optional[TLOCListEntry] = Field(
        default=None, validation_alias="tlocList", serialization_alias="tlocList"
    )

    @model_validator(mode="after")
    def tloc_xor_tloc_list(self):
        check_fields_exclusive(self.__dict__, {"tloc", "tloc_list"}, True)
        return self


class ServiceEntry(BaseModel):
    field: Literal["service"] = "service"
    value: ServiceEntryValue


class TLOCActionEntry(BaseModel):
    field: Literal["tlocAction"] = "tlocAction"
    value: TLOCActionType


class AffinityEntry(BaseModel):
    field: Literal["affinity"] = "affinity"
    value: str = Field(description="Number in range 0-63")


RedirectDNSActionEntry = Union[IPAddressEntry, DNSTypeEntry]


class LogAction(BaseModel):
    type: Literal["log"] = "log"
    parameter: str = ""


class CountAction(BaseModel):
    type: Literal["count"] = "count"
    parameter: str


class NATAction(BaseModel):
    type: Literal["nat"] = "nat"
    parameter: Union[NATPoolEntry, NATVPNEntry]

    @staticmethod
    def from_nat_pool(nat_pool: int) -> "NATAction":
        return NATAction(parameter=NATPoolEntry(value=str(nat_pool)))

    @staticmethod
    def from_nat_vpn(fallback: bool, vpn: int = 0) -> "NATAction":
        return NATAction(parameter=NATVPNEntry.from_nat_vpn(fallback=fallback, vpn=vpn))


class CFlowDAction(BaseModel):
    type: Literal["cflowd"] = "cflowd"


class RedirectDNSAction(BaseModel):
    type: Literal["redirectDns"] = "redirectDns"
    parameter: RedirectDNSActionEntry

    @staticmethod
    def from_ip_address(ip: IPv4Address) -> "RedirectDNSAction":
        return RedirectDNSAction(parameter=IPAddressEntry(value=ip))

    @staticmethod
    def from_dns_type(dns_type: DNSTypeEntryType = "host") -> "RedirectDNSAction":
        return RedirectDNSAction(parameter=DNSTypeEntry(value=dns_type))


class TCPOptimizationAction(BaseModel):
    type: Literal["tcpOptimization"] = "tcpOptimization"
    parameter: str = ""


class DREOptimizationAction(BaseModel):
    type: Literal["dreOptimization"] = "dreOptimization"
    parameter: str = ""


class ServiceNodeGroupAction(BaseModel):
    type: Literal["serviceNodeGroup"] = "serviceNodeGroup"
    parameter: str = Field(default="", pattern=r"^(SNG-APPQOE(3[01]|[12][0-9]|[1-9])?)?$")


class LossProtectionAction(BaseModel):
    type: Literal["lossProtect"] = "lossProtect"
    parameter: LossProtectionType


class LossProtectionFECAction(BaseModel):
    type: Literal["lossProtectFec"] = "lossProtectFec"
    parameter: LossProtectionType = "fecAlways"
    value: Optional[str] = Field(default=None, description="BETA number in range 1-5")


class LossProtectionPacketDuplicationAction(BaseModel):
    type: Literal["lossProtectPktDup"] = "lossProtectPktDup"
    parameter: LossProtectionType = "packetDuplication"


class SecureInternetGatewayAction(BaseModel):
    type: Literal["sig"] = "sig"
    parameter: str = ""


class FallBackToRoutingAction(BaseModel):
    type: Literal["fallbackToRouting"] = "fallbackToRouting"
    parameter: str = ""


class ExportToAction(BaseModel):
    type: Literal["exportTo"] = "exportTo"
    parameter: VPNListEntry


class MirrorAction(BaseModel):
    type: Literal["mirror"] = "mirror"
    parameter: Reference


class ClassMapAction(BaseModel):
    type: Literal["class"] = "class"
    parameter: Reference


class PolicerAction(BaseModel):
    type: Literal["policer"] = "policer"
    parameter: Reference


ActionSetEntry = Annotated[
    Union[
        AffinityEntry,
        CommunityAdditiveEntry,
        CommunityEntry,
        DSCPEntry,
        ForwardingClassEntry,
        LocalTLOCListEntry,
        NextHopEntry,
        NextHopLooseEntry,
        OMPTagEntry,
        PolicerListEntry,
        PreferenceEntry,
        PrefferedColorGroupListEntry,
        ServiceChainEntry,
        ServiceEntry,
        TLOCActionEntry,
        TLOCEntry,
        TLOCListEntry,
        TrafficClassEntry,
        VPNEntry,
    ],
    Field(discriminator="field"),
]


class ActionSet(BaseModel):
    type: Literal["set"] = "set"
    parameter: List[ActionSetEntry] = []


ActionEntry = Annotated[
    Union[
        ActionSet,
        CFlowDAction,
        ClassMapAction,
        CountAction,
        DREOptimizationAction,
        FallBackToRoutingAction,
        LogAction,
        LossProtectionAction,
        LossProtectionFECAction,
        LossProtectionPacketDuplicationAction,
        MirrorAction,
        NATAction,
        PolicerAction,
        RedirectDNSAction,
        SecureInternetGatewayAction,
        ServiceNodeGroupAction,
        TCPOptimizationAction,
        ExportToAction,
    ],
    Field(discriminator="type"),
]

MatchEntry = Annotated[
    Union[
        AppListEntry,
        CarrierEntry,
        ClassMapListEntry,
        ColorListEntry,
        CommunityListEntry,
        DestinationDataIPv6PrefixListEntry,
        DestinationDataPrefixListEntry,
        DestinationFQDNEntry,
        DestinationFQDNListEntry,
        DestinationGeoLocationEntry,
        DestinationGeoLocationListEntry,
        DestinationIPEntry,
        DestinationIPv6Entry,
        DestinationPortEntry,
        DestinationPortListEntry,
        DestinationRegionEntry,
        DNSAppListEntry,
        DNSEntry,
        DomainIDEntry,
        DSCPEntry,
        ExpandedCommunityListEntry,
        GroupIDEntry,
        NextHeaderEntry,
        OMPTagEntry,
        OriginatorEntry,
        OriginEntry,
        PacketLengthEntry,
        PathTypeEntry,
        PLPEntry,
        PreferenceEntry,
        PrefixListEntry,
        ProtocolEntry,
        ProtocolNameEntry,
        ProtocolNameListEntry,
        RegionEntry,
        RegionListEntry,
        RoleEntry,
        RuleSetListEntry,
        SiteEntry,
        SiteListEntry,
        SiteListEntry,
        SourceDataIPv6PrefixListEntry,
        SourceDataPrefixListEntry,
        SourceFQDNEntry,
        SourceFQDNListEntry,
        SourceGeoLocationEntry,
        SourceGeoLocationListEntry,
        SourceIPEntry,
        SourceIPv6Entry,
        SourcePortEntry,
        SourcePortListEntry,
        TCPEntry,
        TLOCEntry,
        TLOCListEntry,
        TrafficClassEntry,
        TrafficToEntry,
        VPNListEntry,
    ],
    Field(discriminator="field"),
]

MUTUALLY_EXCLUSIVE_FIELDS = [
    {"destinationDataPrefixList", "destinationIp"},
    {"destinationDataIpv6PrefixList", "destinationIpv6"},
    {"sourceDataPrefixList", "sourceIp"},
    {"sourceDataIpv6PrefixList", "sourceIpv6"},
    {"protocolName", "protocolNameList", "protocol", "destinationPort", "destinationPortList"},
    {"localTlocList", "preferredColorGroup"},
    {"sig", "fallbackToRouting", "nat", "nextHop", "serviceChain"},
    {"regionId", "regionList"},
    {"siteId", "siteList"},
    {"tloc", "tlocList"},
    {"service", "tlocAction"},
]


def _generate_field_name_check_lookup(spec: Sequence[Set[str]]) -> Dict[str, List[str]]:
    lookup: Dict[str, List[str]] = {}
    for exclusive_set in spec:
        for field in exclusive_set:
            lookup[field] = list(exclusive_set - {field})
    return lookup


MUTUALLY_EXCLUSIVE_FIELD_LOOKUP = _generate_field_name_check_lookup(MUTUALLY_EXCLUSIVE_FIELDS)


class Match(BaseModel):
    entries: Sequence[MatchEntry]


class Action(BaseModel):
    pass


class PolicyDefinitionSequenceBase(BaseModel):
    sequence_id: int = Field(default=0, serialization_alias="sequenceId", validation_alias="sequenceId")
    sequence_name: str = Field(serialization_alias="sequenceName", validation_alias="sequenceName")
    base_action: PolicyActionType = Field(
        default="drop", serialization_alias="baseAction", validation_alias="baseAction"
    )
    sequence_type: SequenceType = Field(serialization_alias="sequenceType", validation_alias="sequenceType")
    sequence_ip_type: SequenceIpType = Field(serialization_alias="sequenceIpType", validation_alias="sequenceIpType")
    ruleset: Optional[bool] = None
    match: Match
    actions: Sequence[ActionEntry]

    @staticmethod
    def _check_field_collision(field: str, fields: Sequence[str]) -> None:
        existing_fields = set(fields)
        forbidden_fields = set(MUTUALLY_EXCLUSIVE_FIELD_LOOKUP.get(field, []))
        colliding_fields = set(existing_fields) & set(forbidden_fields)
        assert not colliding_fields, f"{field} is mutually exclusive with {colliding_fields}"

    def _check_match_can_be_inserted(self, match: MatchEntry) -> None:
        self._check_field_collision(
            match.field,
            [entry.field for entry in self.match.entries],
        )

    def _check_action_can_be_inserted_in_set(
        self, action: ActionSetEntry, action_set_param: List[ActionSetEntry]
    ) -> None:
        self._check_field_collision(
            action.field,
            [param.field for param in action_set_param],
        )

    def _get_match_entries_by_field(self, field: str) -> Sequence[MatchEntry]:
        return [entry for entry in self.match.entries if entry.field == field]

    def _remove_match(self, match_type: Any) -> None:
        if isinstance(self.match.entries, MutableSequence):
            self.match.entries[:] = [entry for entry in self.match.entries if type(entry) != match_type]

    def _insert_match(self, match: MatchEntry, insert_field_check: bool = True) -> int:
        # inserts new item or replaces item with same field name if found
        if insert_field_check:
            self._check_match_can_be_inserted(match)
        if isinstance(self.match.entries, MutableSequence):
            for index, entry in enumerate(self.match.entries):
                if match.field == entry.field:
                    self.match.entries[index] = match
                    return index
            self.match.entries.append(match)
            return len(self.match.entries) - 1
        else:
            raise TypeError("Match entries must be defined as MutableSequence (eg. List) to use _insert_match method")

    def _insert_action(self, action: ActionEntry) -> None:
        if isinstance(self.actions, MutableSequence):
            for index, entry in enumerate(self.actions):
                if action.type == entry.type:
                    self.actions[index] = action
                    return
            self.actions.append(action)
        else:
            raise TypeError("Action entries must be defined as MutableSequence (eg. List) to use _insert_match method")

    def _remove_action(self, action_type_name: str) -> None:
        if isinstance(self.actions, MutableSequence):
            self.actions[:] = [action for action in self.actions if action.type != action_type_name]

    def _insert_action_in_set(self, action: ActionSetEntry) -> None:
        if isinstance(self.actions, MutableSequence):
            # Check if ActionSet entry already exist
            action_sets = [act for act in self.actions if isinstance(act, ActionSet)]
            if len(action_sets) < 1:
                # if not found insert new empty ActionSet
                action_set = ActionSet()
                self.actions.append(action_set)
            else:
                action_set = action_sets[0]
            # Now we operate on action_set parameter list
            self._check_action_can_be_inserted_in_set(action, action_set.parameter)
            for index, param in enumerate(action_set.parameter):
                if action.field == param.field:
                    action_set.parameter[index] = action
                    return
            action_set.parameter.append(action)

    def _remove_action_from_set(self, field_name: str) -> None:
        if isinstance(self.actions, MutableSequence):
            for action in self.actions:
                if isinstance(action, ActionSet):
                    action.parameter[:] = [param for param in action.parameter if param.field != field_name]


def accept_action(method):
    @wraps(method)
    def wrapper(self: PolicyDefinitionSequenceBase, *args, **kwargs):
        assert self.base_action == "accept", f"{method.__name__} only allowed when base_action is accept"
        return method(self, *args, **kwargs)

    return wrapper


class DefaultAction(BaseModel):
    type: PolicyActionType


class InfoTag(BaseModel):
    info_tag: Optional[str] = Field("", serialization_alias="infoTag", validation_alias="infoTag")


class PolicyDefinitionId(BaseModel):
    definition_id: UUID = Field(serialization_alias="definitionId", validation_alias="definitionId")


class PolicyReference(BaseModel):
    id: UUID
    property: str


class DefinitionWithSequencesCommonBase(BaseModel):
    default_action: Optional[DefaultAction] = Field(
        default=DefaultAction(type="drop"),
        serialization_alias="defaultAction",
        validation_alias="defaultAction",
    )
    sequences: Optional[Sequence[PolicyDefinitionSequenceBase]] = None

    def _enumerate_sequences(self, from_index: int = 0) -> None:
        """Updates sequence entries with appropriate index.

        Args:
            from_index (int, optional): Only rules after that index in table will be updated. Defaults to 0.
        """
        if isinstance(self.sequences, MutableSequence):
            start_index = from_index
            sequence_count = len(self.sequences)
            if from_index < 0:
                start_index = sequence_count - start_index
            for i in range(start_index, sequence_count):
                self.sequences[i].sequence_id = i + 1
        else:
            raise TypeError("sequences be defined as MutableSequence (eg. List) to use _enumerate_sequences method")

    def pop(self, index: int = -1) -> None:
        """Removes a sequence item at given index, consecutive sequence items will be enumarated again.

        Args:
            index (int, optional): Defaults to -1.
        """
        if isinstance(self.sequences, MutableSequence):
            self.sequences.pop(index)
            self._enumerate_sequences(index)
        else:
            raise TypeError("sequences be defined as MutableSequence (eg. List) to use pop method")

    def add(self, item: PolicyDefinitionSequenceBase) -> int:
        """Adds new sequence item as last in table, index will be autogenerated.

        Args:
            item (DefinitionSequence): item to be added to sequences

        Returns:
            int: index at which item was added
        """
        if isinstance(self.sequences, MutableSequence):
            insert_index = len(self.sequences)
            self.sequences.append(item)
            self._enumerate_sequences(insert_index)
            return insert_index
        else:
            raise TypeError("sequences be defined as MutableSequence (eg. List) to add method")


class PolicyDefinitionBase(BaseModel):
    name: str = Field(
        pattern="^[a-zA-Z0-9_-]{1,128}$",
        description="Can include only alpha-numeric characters, hyphen '-' or underscore '_'; maximum 128 characters",
    )
    description: str = "default description"
    type: str
    mode: Optional[str] = None
    optimized: Optional[Optimized] = "false"


class PolicyDefinitionInfo(PolicyDefinitionBase, PolicyDefinitionId):
    last_updated: datetime.datetime = Field(serialization_alias="lastUpdated", validation_alias="lastUpdated")
    owner: str
    reference_count: int = Field(serialization_alias="referenceCount", validation_alias="referenceCount")
    references: List[PolicyReference]


class PolicyDefinitionGetResponse(PolicyDefinitionInfo):
    is_activated_by_vsmart: bool = Field(
        serialization_alias="isActivatedByVsmart", validation_alias="isActivatedByVsmart"
    )


class PolicyDefinitionEditResponse(BaseModel):
    master_templates_affected: List[str] = Field(
        default=[], serialization_alias="masterTemplatesAffected", validation_alias="masterTemplatesAffected"
    )


class PolicyDefinitionPreview(BaseModel):
    preview: str


class PolicyDefinitionEndpoints(Protocol):
    def create_policy_definition(self, payload: BaseModel) -> PolicyDefinitionId:
        ...

    def delete_policy_definition(self, id: UUID) -> None:
        ...

    def edit_policy_definition(self, id: UUID, payload: BaseModel) -> PolicyDefinitionEditResponse:
        ...

    def get_definitions(self) -> DataSequence[PolicyDefinitionInfo]:
        ...

    def get_policy_definition(self, id: UUID) -> PolicyDefinitionGetResponse:
        ...



# Copyright 2022 Cisco Systems, Inc. and its affiliates

from __future__ import annotations

import logging
import time
from contextlib import contextmanager
from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING, Iterator, Optional

from catalystwan.api.basic_api import DeviceStateAPI
from catalystwan.dataclasses import Device, PacketSetup, Status
from catalystwan.utils.creation_tools import create_dataclass

if TYPE_CHECKING:
    from catalystwan.session import ManagerSession

logger = logging.getLogger(__name__)


class DownloadStatus(Enum):
    COMPLETED = "COMPLETED"
    IMPOSSIBLE = "IMPOSSIBLE"
    FILESIZE = None


class PacketCaptureAPI:
    """Methods for setting up packet capture session, and download .pcap session file.

    Args:
        session:
        vpn: Defaults to 0.
        interface: Defaults to ge0/1.

    Examples:
        >>> edge_device = session.api.devices.get().filter(personality=Personality.EDGE)[0]
        >>> packet_capture = session.api.packet_capture.get_packets(edge_device)
    """

    def __init__(
        self,
        session: ManagerSession,
        vpn: str = "0",
        interface: str = "ge0/1",
    ) -> None:
        self.session = session
        self.vpn = vpn
        self.interface = interface

    def get_packets(self, device: Device, duration_seconds=120) -> Status:
        """Initiate packet capture process.

        Args:
            device (Device): Device class object
            duration_seconds (int, optional): Duration od packet capturing . Defaults to 20.

        Returns:
            Status
        """
        with DeviceStateAPI(self.session).enable_data_stream():
            try:
                with self.channel(device):
                    with self.start_stop(device):
                        time.sleep(duration_seconds)
            except PermissionError as err:
                failed_status_dict = {
                    "fileDownloadStatus": DownloadStatus.IMPOSSIBLE.value,
                    "fileSize": DownloadStatus.FILESIZE.value,
                }
                self.status = create_dataclass(Status, failed_status_dict)
                logger.error(str(err))
            return self.status

    @contextmanager
    def channel(self, device: Device) -> Iterator:
        """Creates packet capture session.

        Args:
            device (Device): Device class object

        Raises:
            PermissionError: if already another packet capture session is active

        Yields:
            Iterator: packet_channel which contains session ID
        """
        query = {
            "deviceUUID": device.uuid,
            "vpn": self.vpn,
            "interface": self.interface,
            "type": "control",
        }

        try:
            url_path = r"/dataservice/stream/device/capture"
            packet_setup = self.session.post(url=url_path, json=query).json()
            logger.debug(f"Packet capture session for device {device.uuid} has been opened")
            self.packet_channel = create_dataclass(PacketSetup, packet_setup)
            if self.packet_channel.is_new_session is True:
                yield self.packet_channel
            else:
                self.status = None  # type: ignore
                raise PermissionError("Can't start new session, another is already open")
        finally:
            for _ in range(3):
                time.sleep(10)
                self.status = self.get_status(self.packet_channel)
                if self.status.file_download_status == DownloadStatus.COMPLETED.value:
                    self.download_capture_session(self.packet_channel, device)
                    logger.debug(f"Packet downloading for device {device.uuid} has been finished")
                    url_path = f"/dataservice/stream/device/capture/disable/{self.packet_channel.session_id}"
                    self.session.get_json(url_path)
                    logger.debug(f"Packet capture session for device {device.uuid} has been disabled")
                    break

    @contextmanager
    def start_stop(self, device: Device) -> Iterator:
        """Start and stops packet capturing.

        Yields:
            Iterator: None
        """
        try:
            url_path = f"/dataservice/stream/device/capture/start/{self.packet_channel.session_id}"
            self.session.get_json(url_path)
            logger.debug(f"Packet capturing for device {device.uuid} has been started")
            yield None

        finally:
            url_path = f"/dataservice/stream/device/capture/stop/{self.packet_channel.session_id}"
            self.session.get_json(url_path)
            logger.debug(f"Packet capturing for device {device.uuid} has been stoped")

    def get_interface_name(self, device: Device) -> str:
        url_path = f"/dataservice/device/interface/synced?deviceId={device.local_system_ip}"
        ifname = dict(self.session.get_json(url_path))
        if_name = str(ifname["data"][0]["ifname"])
        return if_name

    def download_capture_session(self, packet: PacketSetup, device: Device, file_path: Optional[str] = None) -> bool:
        if file_path is None:
            file_path = f"{Path(__file__).parents[0]}/{device.uuid}.pcap"
        url = f"/dataservice/stream/device/capture/download/{packet.session_id}"
        self.session.get_file(url, file_path)  # type: ignore
        return True

    def get_status(self, packet_channel: PacketSetup) -> Status:
        url_path = f"/dataservice/stream/device/capture/status/{packet_channel.session_id}"
        status = self.session.get_json(url_path)
        self.status = create_dataclass(Status, status)
        return self.status



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from typing import List, Optional, Union
from uuid import UUID

from pydantic import BaseModel, ConfigDict, Field

from catalystwan.api.configuration_groups.parcel import Default, Global, Variable
from catalystwan.models.configuration.feature_profile.sdwan.service.lan.common import (
    Arp,
    StaticIPv4Address,
    StaticIPv6Address,
    VrrpIPv6Address,
    VrrpTrackingObject,
)


class VrrpIPv4SecondaryAddress(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    address: Union[Variable, Global[str]]


class VrrpIPv6SecondaryAddress(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    prefix: Union[Global[str], Variable]


class VrrpIPv4(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    group_id: Union[Variable, Global[int]] = Field(serialization_alias="groupId", validation_alias="groupId")
    priority: Union[Variable, Global[int], Default[int]] = Default[int](value=100)
    timer: Union[Variable, Global[int], Default[int]] = Default[int](value=1000)
    track_omp: Union[Global[bool], Default[bool]] = Field(
        serialization_alias="trackOmp", validation_alias="trackOmp", default=Default[bool](value=False)
    )
    ip_address: Union[Global[str], Variable] = Field(serialization_alias="ipAddress", validation_alias="ipAddress")
    ip_address_secondary: Optional[List[VrrpIPv4SecondaryAddress]] = Field(
        serialization_alias="ipAddressSecondary", validation_alias="ipAddressSecondary"
    )
    tloc_pref_change: Union[Global[bool], Default[bool]] = Field(
        serialization_alias="tlocPrefChange", validation_alias="tlocPrefChange", default=Default[bool](value=False)
    )
    tloc_pref_change_value: Optional[Union[Global[int], Variable]] = Field(
        serialization_alias="tlocPrefChangeValue", validation_alias="tlocPrefChangeValue", default=None
    )
    tracking_object: Optional[List[VrrpTrackingObject]] = Field(
        serialization_alias="trackingObject", validation_alias="trackingObject", default=None
    )

    prefix_list: Optional[Union[Global[str], Variable, Default[None]]] = Default[None](value=None)


class VrrpIPv6(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    group_id: Union[Variable, Global[int]] = Field(serialization_alias="groupId", validation_alias="groupId")
    priority: Union[Variable, Global[int], Default[int]] = Default[int](value=100)
    timer: Union[Variable, Global[int], Default[int]] = Default[int](value=1000)
    track_omp: Union[Global[bool], Default[bool]] = Field(
        serialization_alias="trackOmp", validation_alias="trackOmp", default=Default[bool](value=False)
    )
    track_prefix_list: Optional[Union[Global[str], Variable, Default[None]]] = Field(
        serialization_alias="trackPrefixList", validation_alias="trackPrefixList"
    )
    ipv6: List[VrrpIPv6Address]
    ipv6_secondary: Optional[List[VrrpIPv6SecondaryAddress]]


class Dhcpv6Helper(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    address: Union[Global[str], Variable] = Field(serialization_alias="address", validation_alias="address")
    vpn: Optional[Union[Global[int], Variable, Default[None]]] = None


class AdvancedSviAttributes(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    tcp_mss: Optional[Union[Global[int], Variable, Default[None]]] = Field(
        serialization_alias="tcpMss", validation_alias="tcpMss", default=Default[None](value=None)
    )
    arp_timeout: Optional[Union[Global[int], Variable, Default[int]]] = Field(
        serialization_alias="arpTimeout", validation_alias="arpTimeout", default=Default[int](value=1200)
    )
    ip_directed_broadcast: Optional[Union[Global[bool], Variable, Default[bool]]] = Field(
        serialization_alias="ipDirectedBroadcast",
        validation_alias="ipDirectedBroadcast",
        default=Default[bool](value=False),
    )
    icmp_redirect_disable: Optional[Union[Global[bool], Variable, Default[bool]]] = Field(
        serialization_alias="icmpRedirectDisable",
        validation_alias="icmpRedirectDisable",
        default=Default[bool](value=True),
    )


class IPv4Address(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    address: StaticIPv4Address = Field(serialization_alias="addressV4", validation_alias="addressV4")
    secondary_address: Optional[List[StaticIPv4Address]] = Field(
        serialization_alias="secondaryAddressV4", validation_alias="secondaryAddressV4", default=None
    )
    dhcp_helper: Optional[Union[Global[List[str]], Variable, Default[None]]] = Field(
        serialization_alias="dhcpHelperV4", validation_alias="dhcpHelperV4", default=None
    )


class IPv6Address(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    address: Union[Global[str], Variable, Default[None]] = Field(
        serialization_alias="addressV6", validation_alias="addressV6"
    )
    secondary_address: Optional[List[StaticIPv6Address]] = Field(
        serialization_alias="secondaryAddressV6", validation_alias="secondaryAddressV6", default=None
    )
    dhcp_helper: Optional[List[Dhcpv6Helper]] = Field(
        serialization_alias="dhcpHelperV6", validation_alias="dhcpHelperV6", default=None
    )


class AclQos(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    ipv4_acl_egress: Optional[Global[UUID]] = Field(
        serialization_alias="ipv4AclEgress", validation_alias="ipv4AclEgress", default=None
    )
    ipv4_acl_ingress: Optional[Global[UUID]] = Field(
        serialization_alias="ipv4AclIngress", validation_alias="ipv4AclIngress", default=None
    )
    ipv6_acl_egress: Optional[Global[UUID]] = Field(
        serialization_alias="ipv6AclEgress", validation_alias="ipv6AclEgress", default=None
    )
    ipv6_acl_ingress: Optional[Global[UUID]] = Field(
        serialization_alias="ipv6AclIngress", validation_alias="ipv6AclIngress", default=None
    )


class InterfaceSviData(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    shutdown: Union[Global[bool], Variable, Default[bool]] = Default[bool](value=True)
    interface_name: Optional[Union[Global[str], Variable]] = Field(
        serialization_alias="interfaceName", validation_alias="interfaceName"
    )
    description: Optional[Union[Global[str], Variable, Default[None]]] = Default[None](value=None)
    interface_mtu: Optional[Union[Global[int], Variable, Default[int]]] = Field(
        serialization_alias="ifMtu", validation_alias="ifMtu", default=Default[int](value=1500)
    )
    ip_mtu: Optional[Union[Global[int], Variable, Default[int]]] = Field(
        serialization_alias="ipMtu", validation_alias="ipMtu", default=Default[int](value=1500)
    )
    ipv4: Optional[IPv4Address] = None
    ipv6: Optional[IPv6Address] = None
    acl_qos: Optional[AclQos] = Field(serialization_alias="aclQos", validation_alias="aclQos", default=None)
    arp: Optional[List[Arp]] = None
    vrrp: Optional[List[VrrpIPv4]] = None
    vrrp_ipv6: Optional[List[VrrpIPv6]] = Field(
        serialization_alias="vrrpIpv6", validation_alias="vrrpIpv6", default=None
    )
    dhcp_client_v6: Optional[Union[Global[bool], Variable, Default[bool]]] = Field(
        serialization_alias="dhcpClientV6", validation_alias="dhcpClientV6", default=Default[bool](value=False)
    )
    advanced: AdvancedSviAttributes = AdvancedSviAttributes()


class InterfaceSviCreationPayload(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    name: str
    description: Optional[str] = None
    data: InterfaceSviData
    metadata: Optional[dict] = None



# Copyright 2023 Cisco Systems, Inc. and its affiliates

# mypy: disable-error-code="empty-body"
from enum import Enum
from typing import Optional

from pydantic import BaseModel, ConfigDict, Field

from catalystwan.api.configuration_groups.parcel import _ParcelBase
from catalystwan.endpoints import JSON, APIEndpoints, delete, get, post, put, versions
from catalystwan.models.configuration.feature_profile.common import (
    FeatureProfileCreationPayload,
    FeatureProfileCreationResponse,
    FeatureProfileInfo,
)
from catalystwan.models.feature_profile_parcel import FullConfigParcel
from catalystwan.typed_list import DataSequence


class SchemaType(str, Enum):
    POST = "post"
    PUT = "put"


class SchemaTypeQuery(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    schema_type: SchemaType = Field(alias="schemaType")


class ParcelId(BaseModel):
    id: str = Field(alias="parcelId")


class GetFeatureProfilesPayload(BaseModel):
    limit: Optional[int]
    offset: Optional[int]


class ConfigurationFeatureProfile(APIEndpoints):
    @versions(supported_versions=(">=20.9"), raises=False)
    @get("/v1/feature-profile/sdwan/system/aaa/schema", resp_json_key="request")
    def get_sdwan_system_aaa_parcel_schema(self, params: SchemaTypeQuery) -> JSON:
        ...

    @versions(supported_versions=(">=20.9"), raises=False)
    @get("/v1/feature-profile/sdwan/transport/cellular-controller/schema", resp_json_key="request")
    def get_sdwan_transport_cellular_controller_parcel_schema(self, params: SchemaTypeQuery) -> JSON:
        ...

    @versions(supported_versions=(">=20.9"), raises=False)
    @get("/v1/feature-profile/sdwan/transport")
    def get_sdwan_transport_feature_profiles(
        self, payload: Optional[GetFeatureProfilesPayload]
    ) -> DataSequence[FeatureProfileInfo]:
        ...

    @versions(supported_versions=(">=20.9"), raises=False)
    @post("/v1/feature-profile/sdwan/transport")
    def create_sdwan_transport_feature_profile(
        self, payload: FeatureProfileCreationPayload
    ) -> FeatureProfileCreationResponse:
        ...

    @versions(supported_versions=(">=20.9"), raises=False)
    @delete("/v1/feature-profile/sdwan/transport/{transport_id}")
    def delete_sdwan_transport_feature_profile(self, transport_id: str) -> None:
        ...

    @versions(supported_versions=(">=20.9"), raises=False)
    @get("/v1/feature-profile/sdwan/system")
    def get_sdwan_system_feature_profiles(
        self, payload: Optional[GetFeatureProfilesPayload]
    ) -> DataSequence[FeatureProfileInfo]:
        ...

    @versions(supported_versions=(">=20.9"), raises=False)
    @post("/v1/feature-profile/sdwan/system")
    def create_sdwan_system_feature_profile(
        self, payload: FeatureProfileCreationPayload
    ) -> FeatureProfileCreationResponse:
        ...

    @versions(supported_versions=(">=20.9"), raises=False)
    @delete("/v1/feature-profile/sdwan/system/{system_id}")
    def delete_sdwan_system_feature_profile(self, system_id: str) -> None:
        ...

    @versions(supported_versions=(">=20.9"), raises=False)
    @post("/v1/feature-profile/sdwan/system/{system_id}/aaa")
    def create_aaa_profile_parcel_for_system(self, system_id: str, payload: _ParcelBase) -> ParcelId:
        ...

    @versions(supported_versions=(">=20.9"), raises=False)
    @post("/v1/feature-profile/sdwan/transport/{transport_id}/cellular-controller")
    def create_cellular_controller_profile_parcel_for_transport(
        self, transport_id: str, payload: _ParcelBase
    ) -> ParcelId:
        ...

    @versions(supported_versions=(">=20.9"), raises=False)
    @get("/v1/feature-profile/sdwan")
    def get_sdwan_feature_profiles(self) -> DataSequence[FeatureProfileInfo]:
        ...


class SDRoutingConfigurationFeatureProfile(APIEndpoints):
    @versions(supported_versions=(">=20.13"), raises=False)
    @post("/v1/feature-profile/sd-routing/cli")
    def create_cli_feature_profile(self, payload: FeatureProfileCreationPayload) -> FeatureProfileCreationResponse:
        ...

    @versions(supported_versions=(">=20.13"), raises=False)
    @post("/v1/feature-profile/sd-routing/cli/{cli_fp_id}/full-config")
    def create_cli_full_config_parcel(self, cli_fp_id: str, payload: FullConfigParcel) -> ParcelId:
        ...

    @versions(supported_versions=(">=20.13"), raises=False)
    @delete("/v1/feature-profile/sd-routing/cli/{cli_fp_id}")
    def delete_cli_feature_profile(self, cli_fp_id: str) -> None:
        ...

    @versions(supported_versions=(">=20.13"), raises=False)
    @delete("/v1/feature-profile/sd-routing/cli/{cli_fp_id}/full-config/{parcel_id}")
    def delete_cli_full_config_parcel(self, cli_fp_id: str, parcel_id: str) -> None:
        ...

    @versions(supported_versions=(">=20.13"), raises=False)
    @put("/v1/feature-profile/sd-routing/cli/{cli_fp_id}/full-config/{parcel_id}")
    def edit_cli_full_config_parcel(self, cli_fp_id: str, parcel_id: str, payload: FullConfigParcel) -> None:
        ...

    @versions(supported_versions=(">=20.13"), raises=False)
    @get("/v1/feature-profile/sd-routing/cli")
    def get_cli_feature_profiles(
        self, payload: Optional[GetFeatureProfilesPayload]
    ) -> DataSequence[FeatureProfileInfo]:
        ...



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from enum import Enum
from pathlib import Path
from typing import ClassVar, List, Optional

from pydantic import ConfigDict, Field

from catalystwan.api.templates.bool_str import BoolStr
from catalystwan.api.templates.feature_template import FeatureTemplate, FeatureTemplateValidator


class Version(str, Enum):
    TLSV11 = "TLSv1.1"
    TLSV12 = "TLSv1.2"


class AuthType(str, Enum):
    SERVER = "Server"
    MUTUAL = "Mutual"


class TlsProfile(FeatureTemplateValidator):
    profile: str
    version: Optional[Version] = Field(Version.TLSV11, json_schema_extra={"data_path": ["tls-version"]})
    auth_type: AuthType = Field(json_schema_extra={"vmanage_key": "auth-type"})
    ciphersuite_list: Optional[List] = Field(
        default=None, json_schema_extra={"data_path": ["ciphersuite"], "vmanage_key": "ciphersuite-list"}
    )
    model_config = ConfigDict(populate_by_name=True)


class Priority(str, Enum):
    INFORMATION = "information"
    DEBUGGING = "debugging"
    NOTICE = "notice"
    WARN = "warn"
    ERROR = "error"
    CRITICAL = "critical"
    ALERT = "alert"
    EMERGENCY = "emergency"


class Server(FeatureTemplateValidator):
    name: str
    vpn: Optional[int] = None
    source_interface: Optional[str] = Field(default=None, json_schema_extra={"vmanage_key": "source-interface"})
    priority: Optional[Priority] = Priority.INFORMATION
    enable_tls: Optional[BoolStr] = Field(
        default=False, json_schema_extra={"data_path": ["tls"], "vmanage_key": "enable-tls"}
    )
    custom_profile: Optional[BoolStr] = Field(
        default=False, json_schema_extra={"data_path": ["tls", "tls-properties"], "vmanage_key": "custom-profile"}
    )
    profile: Optional[str] = Field(default=None, json_schema_extra={"data_path": ["tls", "tls-properties"]})
    model_config = ConfigDict(populate_by_name=True)


class Ipv6Server(FeatureTemplateValidator):
    name: str
    vpn: Optional[int] = None
    source_interface: Optional[str] = Field(default=None, json_schema_extra={"vmanage_key": "source-interface"})
    priority: Optional[Priority] = Priority.INFORMATION
    enable_tls: Optional[BoolStr] = Field(
        default=False, json_schema_extra={"data_path": ["tls"], "vmanage_key": "enable-tls"}
    )
    custom_profile: Optional[BoolStr] = Field(
        default=False, json_schema_extra={"data_path": ["tls", "tls-properties"], "vmanage_key": "custom-profile"}
    )
    profile: Optional[str] = Field(default=None, json_schema_extra={"data_path": ["tls", "tls-properties"]})
    model_config = ConfigDict(populate_by_name=True)


class CiscoLoggingModel(FeatureTemplate):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    enable: Optional[BoolStr] = Field(default=None, json_schema_extra={"data_path": ["disk"]})
    size: Optional[int] = Field(default=None, json_schema_extra={"data_path": ["disk", "file"]})
    rotate: Optional[int] = Field(default=None, json_schema_extra={"data_path": ["disk", "file"]})
    tls_profile: Optional[List[TlsProfile]] = Field(default=None, json_schema_extra={"vmanage_key": "tls-profile"})
    server: Optional[List[Server]] = None
    ipv6_server: Optional[List[Ipv6Server]] = Field(default=None, json_schema_extra={"vmanage_key": "ipv6-server"})

    payload_path: ClassVar[Path] = Path(__file__).parent / "DEPRECATED"
    type: ClassVar[str] = "cisco_logging"



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from typing import List

from pydantic import AliasPath, BaseModel, ConfigDict, Field

from catalystwan.api.configuration_groups.parcel import Global, _ParcelBase, as_global


class ProtocolListEntry(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    protocol: Global[str] = Field(serialization_alias="protocolName", validation_alias="protocolName")


class ProtocolListParcel(_ParcelBase):
    entries: List[ProtocolListEntry] = Field(default=[], validation_alias=AliasPath("data", "entries"))

    def add_protocol(self, protocol: str):
        self.entries.append(ProtocolListEntry(protocol=as_global(protocol)))



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from typing import List

from pydantic import AliasPath, BaseModel, ConfigDict, Field

from catalystwan.api.configuration_groups.parcel import Global, _ParcelBase, as_global
from catalystwan.models.common import WellKnownBGPCommunities


class StandardCommunityEntry(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    standard_community: Global[WellKnownBGPCommunities] = Field(
        serialization_alias="standardCommunity", validation_alias="standardCommunity"
    )


class StandardCommunityParcel(_ParcelBase):
    entries: List[StandardCommunityEntry] = Field(default=[], validation_alias=AliasPath("data", "entries"))

    def add_community(self, standard_community: WellKnownBGPCommunities):
        self.entries.append(
            StandardCommunityEntry(standard_community=as_global(standard_community, WellKnownBGPCommunities))
        )



# Copyright 2022 Cisco Systems, Inc. and its affiliates

import unittest
from unittest.mock import patch

from parameterized import parameterized  # type: ignore

from catalystwan.api.task_status_api import SubTaskData, TaskResult
from catalystwan.api.template_api import TemplatesAPI
from catalystwan.api.templates.cli_template import CLITemplate
from catalystwan.api.templates.device_template.device_template import DeviceTemplate
from catalystwan.api.templates.feature_template import FeatureTemplate
from catalystwan.api.templates.models.cisco_aaa_model import CiscoAAAModel
from catalystwan.api.templates.payloads.aaa.aaa_model import AAAModel, AuthenticationOrder
from catalystwan.dataclasses import Device, FeatureTemplateInfo, TemplateInfo
from catalystwan.typed_list import DataSequence
from catalystwan.utils.creation_tools import create_dataclass
from catalystwan.utils.device_model import DeviceModel
from catalystwan.utils.personality import Personality
from catalystwan.utils.reachability import Reachability


class TestTemplatesAPI(unittest.TestCase):
    def setUp(self):
        self.data_template = [
            {
                "deviceType": "vedge-C8000V",
                "lastUpdatedBy": "user",
                "resourceGroup": "global",
                "templateClass": "cedge",
                "configType": "file",
                "templateId": "dummy_id_1",
                "factoryDefault": False,
                "templateName": "template_1",
                "devicesAttached": 0,
                "templateDescription": "dummy template 1",
                "draftMode": "Disabled",
                "lastUpdatedOn": 0,
                "templateAttached": 0,
            },
            {
                "deviceType": "vedge-cloud",
                "lastUpdatedBy": "admin",
                "resourceGroup": "global",
                "templateClass": "vedge",
                "configType": "file",
                "templateId": "dummy_id_2",
                "factoryDefault": False,
                "templateName": "template_2",
                "devicesAttached": 1,
                "templateDescription": "dummy template 2",
                "draftMode": "Disabled",
                "lastUpdatedOn": 0,
                "templateAttached": 0,
            },
        ]
        self.templates = DataSequence(
            TemplateInfo,
            [create_dataclass(TemplateInfo, template) for template in self.data_template],
        )
        self.device_info = Device(
            personality=Personality.EDGE,
            uuid="dummy_uuid",
            id="162.168.0.1",
            hostname="dummy_host",
            reachability=Reachability.REACHABLE,
            local_system_ip="192.168.0.1",
            memUsage=1.0,
            connected_vManages=["192.168.0.1"],
            model="vedge-cloud",
            status="normal",
        )
        sub_tasks_data = SubTaskData.parse_obj(
            {
                "status": "Success",
                "statusId": "success",
                "action": "",
                "activity": [],
                "currentActivity": "",
                "actionConfig": "",
                "order": 1,
                "uuid": "",
                "host-name": "",
                "site-id": "",
            }
        )
        self.task = TaskResult(result=True, sub_tasks_data=[sub_tasks_data])

    @patch("catalystwan.response.ManagerResponse")
    @patch("catalystwan.session.ManagerSession")
    def test_templates_success(self, mock_session, mocked_response):
        # Arrange
        mock_session.get.return_value = mocked_response
        mocked_response.dataseq.return_value = self.templates

        # Act
        answer = TemplatesAPI(mock_session).get(FeatureTemplate)

        # Assert
        self.assertEqual(answer, self.templates)

    @patch("catalystwan.response.ManagerResponse")
    @patch("catalystwan.session.ManagerSession")
    def test_templates_get(self, mock_session, mocked_response):
        # Arrange
        mock_session.get_data.return_value = mocked_response
        mocked_response.dataseq.return_value = DataSequence(FeatureTemplateInfo, [])

        # Act
        TemplatesAPI(mock_session).get(FeatureTemplate)

        # Assert
        mock_session.get.assert_called_once_with(url="/dataservice/template/feature", params={"summary": True})

    @parameterized.expand(
        [
            (
                AAAModel(
                    template_name="test_template",
                    template_description="test_description",
                    auth_order=[AuthenticationOrder.LOCAL],
                    auth_fallback=False,
                    auth_disable_audit_logs=True,
                    auth_admin_order=True,
                    auth_disable_netconf_logs=False,
                ),
                "aaa_id",
            ),
            (
                DeviceTemplate(  # type: ignore
                    template_name="test_device_template",
                    template_description="test_device_template_description",
                    device_type=DeviceModel.VEDGE_2000,
                    general_templates=[],
                ),
                "device_template_id",
            ),
            (
                CLITemplate(  # type: ignore
                    template_name="test", template_description="test", device_model=DeviceModel.VEDGE
                ),
                "cli_id",
            ),
            (
                CiscoAAAModel(  # type: ignore
                    template_name="test_aaa_model",
                    template_description="test_aaa_description",
                ),
                "generator_id",
            ),
        ]
    )
    @patch("catalystwan.session.ManagerSession")
    @patch("catalystwan.api.template_api.TemplatesAPI.create_by_generator")
    @patch("catalystwan.api.template_api.TemplatesAPI._create_feature_template")
    @patch("catalystwan.api.template_api.TemplatesAPI._create_device_template")
    @patch("catalystwan.api.template_api.TemplatesAPI._create_cli_template")
    def test_templates_create_success(
        self,
        template,
        template_id,
        mock_create_cli_template,
        mock_create_device_template,
        mock_create_feature_template,
        mock_create_by_generator,
        mock_session,
    ):
        # Arrange
        mock_create_by_generator.return_value = "generator_id"
        mock_create_feature_template.return_value = "aaa_id"
        mock_create_device_template.return_value = "device_template_id"
        mock_create_cli_template.return_value = "cli_id"

        # Act
        templates_api = TemplatesAPI(mock_session)

        # Assert
        self.assertEqual(templates_api.create(template), template_id)

    @parameterized.expand(
        [
            (
                [
                    AAAModel(
                        template_name="test_template",
                        template_description="test_description",
                        auth_order=[AuthenticationOrder.LOCAL],
                        auth_fallback=False,
                        auth_disable_audit_logs=True,
                        auth_admin_order=True,
                        auth_disable_netconf_logs=False,
                    ),
                    DeviceTemplate(  # type: ignore
                        template_name="test_device_template",
                        template_description="test_device_template_description",
                        device_type=DeviceModel.VEDGE_2000,
                        general_templates=[],
                    ),
                    CLITemplate(  # type: ignore
                        template_name="test_cli_template",
                        template_description="test_cli_description",
                        device_model=DeviceModel.VBOND,
                    ),
                    CiscoAAAModel(  # type: ignore
                        template_name="test_aaa_model",
                        template_description="test_aaa_description",
                    ),
                ],
                ["aaa_id", "device_template_id", "cli_id", "generator_id"],
            ),
        ]
    )
    @patch("catalystwan.session.ManagerSession")
    @patch("catalystwan.api.template_api.TemplatesAPI.create_by_generator")
    @patch("catalystwan.api.template_api.TemplatesAPI._create_feature_template")
    @patch("catalystwan.api.template_api.TemplatesAPI._create_device_template")
    @patch("catalystwan.api.template_api.TemplatesAPI._create_cli_template")
    def test_templates_create_list_success(
        self,
        templates,
        template_ids,
        mock_create_cli_template,
        mock_create_device_template,
        mock_create_feature_template,
        mock_create_by_generator,
        mock_session,
    ):
        # Arrange
        mock_create_by_generator.return_value = "generator_id"
        mock_create_feature_template.return_value = "aaa_id"
        mock_create_device_template.return_value = "device_template_id"
        mock_create_cli_template.return_value = "cli_id"

        # Act
        templates_api = TemplatesAPI(mock_session)

        # Assert
        self.assertEqual(templates_api.create(templates), template_ids)

    @patch("catalystwan.session.ManagerSession")
    @patch("catalystwan.api.template_api.TemplatesAPI.create_by_generator")
    @patch("catalystwan.api.template_api.TemplatesAPI._create_feature_template")
    @patch("catalystwan.api.template_api.TemplatesAPI._create_device_template")
    @patch("catalystwan.api.template_api.TemplatesAPI._create_cli_template")
    def test_templates_create_failure(
        self,
        mock_create_cli_template,
        mock_create_device_template,
        mock_create_feature_template,
        mock_create_by_generator,
        mock_session,
    ):
        # Act
        templates_api = TemplatesAPI(mock_session)

        # Assert
        with self.assertRaises(NotImplementedError):
            templates_api.create(None)
        assert not mock_create_cli_template.called
        assert not mock_create_device_template.called
        assert not mock_create_feature_template.called
        assert not mock_create_by_generator.called

    # @patch.object(TemplatesAPI, "templates")
    # @patch("catalystwan.api.template_api.wait_for_completed")
    # @patch("catalystwan.session.ManagerSession")
    # def test_attach_exist_template(self, mock_session, mock_wait_for_completed, mock_templates):

    #     # Arrage
    #     mock_session.post_json.return_value = {"id": "operation_id"}
    #     test_object = TemplatesAPI(mock_session)

    #     # mock templates
    #     MockTemplates = Mock()
    #     mock_templates.return_value = MockTemplates
    #     test_object.templates = self.templates

    #     # mock wait complete
    #     mock_wait_for_completed.return_value = self.task

    #     # Act
    #     answer = test_object.attach("template_1", self.device_info)

    #     # Assert
    #     self.assertTrue(answer)

    # @patch.object(TemplatesAPI, "templates")
    # @patch("catalystwan.api.template_api.wait_for_completed")
    # @patch("catalystwan.session.ManagerSession")
    # def test_attach_no_exist_template(self, mock_session, mock_wait_for_completed, mock_templates):

    #     # Arrage
    #     mock_session.post_json.return_value = {"id": "operation_id"}
    #     test_object = TemplatesAPI(mock_session)

    #     # mock templates
    #     MockTemplates = Mock()
    #     mock_templates.return_value = MockTemplates
    #     test_object.templates = self.templates

    #     # mock wait complete
    #     mock_wait_for_completed.return_value = self.task

    #     # Act
    #     answer = test_object.attach("no_exist_template", self.device_info)

    #     # Assert
    #     self.assertFalse(answer)

    # @patch("catalystwan.api.template_api.wait_for_completed")
    # @patch("catalystwan.session.ManagerSession")
    # def test_device_to_cli_true(self, mock_session, mock_wait_for_completed):

    #     # Arrage
    #     mock_session.post_json.return_value = {"id": "operation_id"}
    #     test_object = TemplatesAPI(mock_session)

    #     # mock wait complete
    #     mock_wait_for_completed.return_value = self.task

    #     # Act
    #     answer = test_object.device_to_cli(self.device_info)

    #     # Assert
    #     self.assertTrue(answer)

    # @patch.object(TemplatesAPI, "templates")
    # @patch("catalystwan.session.ManagerSession")
    # def test_delete_success(self, mock_session, mock_templates):

    #     # Arrage
    #     MockResponse = MagicMock()
    #     MockResponse.status_code = 200
    #     mock_session.delete.return_value = MockResponse
    #     test_object = TemplatesAPI(mock_session)

    #     # mock templates
    #     MockTemplates = Mock()
    #     mock_templates.return_value = MockTemplates
    #     test_object.templates = self.templates

    #     # Act
    #     answer = test_object.delete("template_1")
    #     # Assert
    #     self.assertTrue(answer)

    # @patch.object(TemplatesAPI, "templates")
    # @patch("catalystwan.session.ManagerSession")
    # def test_delete_wrong_status(self, mock_session, mock_templates):

    #     # Arrage
    #     MockResponse = MagicMock()
    #     MockResponse.ok = False
    #     mock_session.delete.return_value = MockResponse
    #     test_object = TemplatesAPI(mock_session)

    #     # mock templates
    #     MockTemplates = Mock()
    #     mock_templates.return_value = MockTemplates
    #     test_object.templates = self.templates

    #     # Act
    #     answer = test_object.delete("template_1")

    #     # Assert
    #     self.assertFalse(answer)

    # @patch.object(TemplatesAPI, "templates")
    # @patch("catalystwan.session.ManagerSession")
    # def test_delete_exception(self, mock_session, mock_templates):

    #     # Arrage
    #     MockResponse = MagicMock()
    #     MockResponse.status = 404
    #     mock_session.delete.return_value = MockResponse
    #     test_object = TemplatesAPI(mock_session)

    #     # mock templates
    #     MockTemplates = Mock()
    #     mock_templates.return_value = MockTemplates
    #     test_object.templates = self.templates

    #     # Act
    #     def answer():
    #         return test_object.delete("template_2")

    #     # Assert
    #     self.assertRaises(AttachedError, answer)

    # @patch.object(TemplatesAPI, "templates")
    # @patch("catalystwan.session.ManagerSession")
    # def test_create_exception(self, mock_session, mock_templates):

    #     # Arrage
    #     test_object = TemplatesAPI(mock_session)

    #     # mock templates
    #     MockTemplates = Mock()
    #     mock_templates.return_value = MockTemplates
    #     test_object.templates = self.templates
    #     config = CiscoConfParse([])

    #     # Act
    #     def answer():
    #         return test_object.create(self.device_info, "template_1", "new_description", config)

    #     # Assert
    #     self.assertRaises(AlreadyExistsError, answer)



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from ipaddress import IPv4Address
from typing import List, Optional

from pydantic import AliasPath, BaseModel, ConfigDict, Field, field_validator

from catalystwan.api.configuration_groups.parcel import Global, _ParcelBase, as_global
from catalystwan.models.common import TLOCColor
from catalystwan.models.policy.lists_entries import EncapType


class TlocEntry(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    tloc: Global[IPv4Address]
    color: Global[TLOCColor]
    encapsulation: Global[EncapType] = Field(serialization_alias="encap", validation_alias="encap")
    preference: Optional[Global[str]] = None

    @field_validator("preference")
    @classmethod
    def ensure_correct_preference_value(cls, v: Global):
        if not v:
            return v
        if not (0 <= int(v.value) < 4_294_967_295):
            raise ValueError('"preference" not in range 0 - 4 294 967 295 (2 ** 32 - 1)')
        return v


class TlocParcel(_ParcelBase):
    entries: List[TlocEntry] = Field(default=[], validation_alias=AliasPath("data", "entries"))

    def add_entry(
        self, tloc: IPv4Address, color: TLOCColor, encapsulation: EncapType, preference: Optional[str] = None
    ):
        self.entries.append(
            TlocEntry(
                tloc=as_global(tloc),
                color=as_global(color, TLOCColor),
                encapsulation=as_global(encapsulation, EncapType),
                preference=as_global(preference) if preference is not None else None,
            )
        )



# Copyright 2023 Cisco Systems, Inc. and its affiliates

#  type: ignore
import json
import unittest
from pathlib import Path
from typing import ClassVar, List, Optional
from unittest import TestCase
from unittest.mock import patch

from parameterized import parameterized
from pydantic import BaseModel, ConfigDict, Field

from catalystwan.api.template_api import TemplatesAPI
from catalystwan.api.templates.feature_template import FeatureTemplate
from catalystwan.session import ManagerSession


class MockedFeatureTemplate(FeatureTemplate):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    template_name: str = "test"
    template_description: str = "test"
    payload_path: ClassVar[Path] = Path(__file__).parent / "DEPRECATED"
    type: ClassVar[str] = "test_type"

    num: Optional[str]


class MockedFeatureTemplateAlias(FeatureTemplate):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    template_name: str = "test"
    template_description: str = "test"
    payload_path: ClassVar[Path] = Path(__file__).parent / "DEPRECATED"
    type: ClassVar[str] = "test_type"

    num: str = Field(alias="as-num")


class RSA(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    key: str = Field(alias="key-string")
    key_type: str = Field(alias="key-type", json_schema_extra={"data_path": ["type", "RSA"]})


class User(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    name: str
    password: str = Field(json_schema_extra={"data_path": ["list"]})
    pubkey_chain: List[RSA] = Field(default=[], alias="pubkey-chain")


class MockedFeatureTemplateChildren(FeatureTemplate):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    template_name: str = "test"
    template_description: str = "test"
    payload_path: ClassVar[Path] = Path(__file__).parent / "DEPRECATED"
    type: ClassVar[str] = "test_type"

    user: List[User]


class DataPathFeatureTemplate(FeatureTemplate):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    template_name: str = "test"
    template_description: str = "test"
    payload_path: ClassVar[Path] = Path(__file__).parent / "DEPRECATED"
    type: ClassVar[str] = "test_type"

    as_num: str = Field(alias="as-num", json_schema_extra={"data_path": ["authentication", "dot1x", "default"]})


password = "pass"  # pragma: allowlist secret

mocked_feature_template_children_1 = MockedFeatureTemplateChildren(
    user=[User(name="user1", password=password), User(name="user2", password=password)]
)

mocked_feature_template_children_2 = MockedFeatureTemplateChildren(
    user=[
        User(name="user1", password=password, pubkey_chain=[RSA(key="*****", key_type="RSA")]),
        User(name="user2", password=password),
    ]
)


class TestFeatureTemplate(TestCase):
    @parameterized.expand(
        [
            ("basic.json", None, MockedFeatureTemplate(num="num")),
            ("alias.json", None, MockedFeatureTemplateAlias(num="12")),
            ("data_path.json", None, DataPathFeatureTemplate(as_num="12")),
            ("children.json", None, mocked_feature_template_children_1),
            ("children_nested.json", None, mocked_feature_template_children_2),
            ("children_nested_datapath.json", None, mocked_feature_template_children_2),
        ]
    )
    @patch("catalystwan.session.ManagerSession")
    def test_get(
        self,
        filename: str,
        definition_name: Optional[str],
        mocked_template: Optional[FeatureTemplate],
        mocked_session: ManagerSession,
    ):
        # Arrange
        templates_api = TemplatesAPI(mocked_session)
        with open(Path(__file__).resolve().parents[0] / Path("schemas") / Path("basic") / Path(filename), "r") as file:
            schema = json.load(file)

        if definition_name is None:
            definition_name = filename

        with open(
            Path(__file__).resolve().parents[0] / Path("definitions") / Path("basic") / Path(definition_name), "r"
        ) as file:
            definition = json.load(file)

        if mocked_template is None:
            mocked_template = MockedFeatureTemplate()

        # Act
        a = templates_api.generate_feature_template_payload(mocked_template, schema).model_dump(
            by_alias=True, exclude_none=True, mode="json"
        )["templateDefinition"]
        print(json.dumps(a))
        # Assert
        self.assertDictEqual(a, definition)


if __name__ == "__main__":
    unittest.main()



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from pathlib import Path
from typing import ClassVar, Optional

from pydantic import ConfigDict, Field

from catalystwan.api.templates.bool_str import BoolStr
from catalystwan.api.templates.feature_template import FeatureTemplate


class OMPvSmart(FeatureTemplate):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    graceful_restart: Optional[BoolStr] = Field(default=None, json_schema_extra={"vmanage_key": "graceful-restart"})
    send_path_limit: Optional[int] = Field(default=None, json_schema_extra={"vmanage_key": "send-path-limit"})
    send_backup_paths: Optional[BoolStr] = Field(default=None, json_schema_extra={"vmanage_key": "send-backup-paths"})
    discard_rejected: Optional[BoolStr] = Field(default=None, json_schema_extra={"vmanage_key": "discard-rejected"})
    shutdown: Optional[BoolStr] = Field(default=None, json_schema_extra={"vmanage_key": "shutdown"})
    graceful_restart_timer: Optional[int] = Field(
        default=None, json_schema_extra={"vmanage_key": "graceful-restart-timer", "data_path": ["timers"]}
    )
    eor_timer: Optional[int] = Field(
        default=None, json_schema_extra={"vmanage_key": "eor-timer", "data_path": ["timers"]}
    )
    holdtime: Optional[int] = Field(
        default=None, json_schema_extra={"vmanage_key": "holdtime", "data_path": ["timers"]}
    )
    affinity_group_preference: Optional[BoolStr] = Field(
        default=None, json_schema_extra={"vmanage_key": "affinity-group-preference"}
    )
    advertisement_interval: Optional[int] = Field(
        default=None, json_schema_extra={"vmanage_key": "advertisement-interval", "data_path": ["timers"]}
    )

    payload_path: ClassVar[Path] = Path(__file__).parent / "DEPRECATED"
    type: ClassVar[str] = "omp-vsmart"
